from itertools import combinations

import mne
import numpy as np

from ..base import BaseSolver, InverseOperator, SolverMeta


def _batch_score(Gram, Q, idx_array):
    """Vectorised score for all subsets: tr(Gram_S^{-1} Q_S)."""
    rows = idx_array[:, :, None]  # (n_subsets, k, 1)
    cols = idx_array[:, None, :]  # (n_subsets, 1, k)
    G_batch = Gram[rows, cols]  # (n_subsets, k, k)
    Q_batch = Q[rows, cols]  # (n_subsets, k, k)
    # solve G_batch @ X = Q_batch  =>  tr(G^{-1} Q) = tr(X)
    X = np.linalg.solve(G_batch, Q_batch)
    return np.trace(X, axis1=-2, axis2=-1)


def _score_single(Gram, Q, subset):
    """Score a single subset."""
    idx = np.array(subset, dtype=np.intp)
    G_sub = Gram[np.ix_(idx, idx)]
    Q_sub = Q[np.ix_(idx, idx)]
    return np.trace(np.linalg.solve(G_sub, Q_sub))


class SolverExhaustiveSubspaceML(BaseSolver):
    """Source localization via exhaustive maximum-likelihood subset search
    with BIC model order selection.

    For k <= k_exhaustive (default 3), evaluates ALL C(n_sources, k) subsets.
    For k > k_exhaustive, uses beam search extending top-B solutions from k-1.

    References
    ----------
    [1] Wax, M., & Kailath, T. (1985). Detection of signals by information
        theoretic criteria. IEEE Trans. ASSP, 33(2), 387-392.
    """

    meta = SolverMeta(
        acronym="ExhaustiveML",
        full_name="Exhaustive Subspace Maximum-Likelihood",
        category="Subspace Methods",
        description=(
            "Source localization via exhaustive maximum-likelihood subset "
            "search (k<=3) with beam search extension (k>3) and BIC model "
            "order selection."
        ),
        references=[
            "Wax, M., & Kailath, T. (1985). Detection of signals by information theoretic criteria. IEEE Trans. ASSP, 33(2), 387-392.",
        ],
    )

    def __init__(self, name="ExhaustiveML", **kwargs):
        self.name = name
        super().__init__(**kwargs)

    def make_inverse_operator(
        self,
        forward,
        mne_obj=None,
        *args,
        alpha="auto",
        noise_cov: mne.Covariance | None = None,
        k_max=5,
        k_exhaustive=3,
        beam_width=50,
        penalty_mode="bic_per_timepoint",
        **kwargs,
    ):
        super().make_inverse_operator(forward, mne_obj, *args, alpha=alpha, **kwargs)
        wf = self.prepare_whitened_forward(noise_cov)

        data = self.unpack_data_obj(mne_obj)
        data_w = wf.sensor_transform @ data  # (n_eff, T)
        G_w = wf.G_white  # (n_eff, n_sources)

        n_eff, n_sources = G_w.shape
        T = data_w.shape[1]

        Gram = G_w.T @ G_w
        # Ridge for numerical stability (rank-deficient when n_sources >> n_eff)
        eps_gram = 1e-10 * np.trace(Gram) / max(n_sources, 1)
        Gram[np.diag_indices_from(Gram)] += eps_gram
        R = G_w.T @ data_w
        Q = R @ R.T
        total_var = np.sum(data_w**2)

        def _penalty(k):
            if penalty_mode == "bic_per_timepoint":
                return k * T * np.log(n_eff)
            elif penalty_mode == "aic":
                return 2 * k * T
            elif penalty_mode == "bic_stacked":
                return k * (T + 1) * np.log(n_eff * T)
            elif penalty_mode == "bic_support_only":
                return k * np.log(n_eff * T)
            else:
                raise ValueError(f"Unknown penalty_mode: {penalty_mode}")

        def _compute_bic(scores, k):
            residuals = np.maximum(total_var - scores, 1e-30)
            return n_eff * T * np.log(residuals / (n_eff * T)) + _penalty(k)

        best_bic = np.inf
        best_subset = None
        chunk_size = 500_000

        top_subsets: list[tuple[int, ...]] = []

        # Limit exhaustive search to feasible sizes (< 2M subsets).
        from math import comb as _comb

        _max_exhaustive_subsets = 2_000_000

        for k in range(1, min(k_max, n_sources) + 1):
            if k <= k_exhaustive and _comb(n_sources, k) <= _max_exhaustive_subsets:
                idx_list = list(combinations(range(n_sources), k))
                idx_array = np.array(idx_list, dtype=np.intp)

                all_score_chunks = []
                for start in range(0, len(idx_array), chunk_size):
                    chunk = idx_array[start : start + chunk_size]
                    scores = _batch_score(Gram, Q, chunk)
                    all_score_chunks.append(scores)

                all_scores = np.concatenate(all_score_chunks)
                bics = _compute_bic(all_scores, k)
                local_best = np.argmin(bics)

                if bics[local_best] < best_bic:
                    best_bic = bics[local_best]
                    best_subset = idx_list[local_best]

                top_indices = np.argsort(all_scores)[-beam_width:]
                top_subsets = [idx_list[i] for i in top_indices]

            else:
                new_subsets: dict[tuple[int, ...], float] = {}
                for subset in top_subsets:
                    subset_set = set(subset)
                    remaining = [j for j in range(n_sources) if j not in subset_set]
                    if not remaining:
                        continue
                    extended = [tuple(sorted(subset + (j,))) for j in remaining]
                    ext_array = np.array(extended, dtype=np.intp)

                    for start in range(0, len(ext_array), chunk_size):
                        chunk = ext_array[start : start + chunk_size]
                        scores = _batch_score(Gram, Q, chunk)
                        for idx, sc in enumerate(scores):
                            sub = extended[start + idx]
                            if sub not in new_subsets or sc > new_subsets[sub]:
                                new_subsets[sub] = sc

                if not new_subsets:
                    break

                ext_list = list(new_subsets.keys())
                ext_scores = np.array(list(new_subsets.values()))
                bics = _compute_bic(ext_scores, k)
                local_best = np.argmin(bics)

                if bics[local_best] < best_bic:
                    best_bic = bics[local_best]
                    best_subset = ext_list[local_best]

                top_indices = np.argsort(ext_scores)[-beam_width:]
                top_subsets = [ext_list[i] for i in top_indices]

        selected_sources = np.array(best_subset, dtype=np.intp)

        G_sel = G_w[:, selected_sources]
        inv_w = np.linalg.lstsq(G_sel, np.eye(n_eff), rcond=None)[0]

        inverse_operator_w = np.zeros((n_sources, n_eff))
        inverse_operator_w[selected_sources, :] = inv_w

        self.inverse_operators = [
            InverseOperator(inverse_operator_w @ wf.sensor_transform, self.name),
        ]
        return self


class SolverGreedyML(BaseSolver):
    """Source localization via multi-start greedy forward selection with
    ML objective, BIC model order selection, and coordinate-wise refinement.

    Scales as O(n_starts * n_sources * k_max) — works for any source space.

    References
    ----------
    [1] Wax, M., & Kailath, T. (1985). Detection of signals by information
        theoretic criteria. IEEE Trans. ASSP, 33(2), 387-392.
    """

    meta = SolverMeta(
        acronym="GreedyML",
        full_name="Greedy Maximum-Likelihood",
        category="Subspace Methods",
        description=(
            "Source localization via multi-start greedy forward selection "
            "with ML objective, BIC model order selection, and coordinate-wise "
            "refinement. Scales to large source spaces."
        ),
        references=[
            "Wax, M., & Kailath, T. (1985). Detection of signals by information theoretic criteria. IEEE Trans. ASSP, 33(2), 387-392.",
        ],
    )

    def __init__(self, name="GreedyML", **kwargs):
        self.name = name
        super().__init__(**kwargs)

    def make_inverse_operator(
        self,
        forward,
        mne_obj=None,
        *args,
        alpha="auto",
        noise_cov: mne.Covariance | None = None,
        k_max=5,
        n_starts=50,
        n_refine_iters=3,
        penalty_mode="bic_per_timepoint",
        **kwargs,
    ):
        super().make_inverse_operator(forward, mne_obj, *args, alpha=alpha, **kwargs)
        wf = self.prepare_whitened_forward(noise_cov)

        data = self.unpack_data_obj(mne_obj)
        data_w = wf.sensor_transform @ data  # (n_eff, T)
        G_w = wf.G_white  # (n_eff, n_sources)

        n_eff, n_sources = G_w.shape
        T = data_w.shape[1]

        Gram = G_w.T @ G_w
        # Ridge for numerical stability (rank-deficient when n_sources >> n_eff)
        eps_gram = 1e-10 * np.trace(Gram) / max(n_sources, 1)
        Gram[np.diag_indices_from(Gram)] += eps_gram
        R = G_w.T @ data_w
        Q = R @ R.T
        total_var = np.sum(data_w**2)

        def _penalty(k):
            if penalty_mode == "bic_per_timepoint":
                return k * T * np.log(n_eff)
            elif penalty_mode == "aic":
                return 2 * k * T
            elif penalty_mode == "bic_stacked":
                return k * (T + 1) * np.log(n_eff * T)
            elif penalty_mode == "bic_support_only":
                return k * np.log(n_eff * T)
            else:
                raise ValueError(f"Unknown penalty_mode: {penalty_mode}")

        def _bic(score, k):
            residual = max(total_var - score, 1e-30)
            return n_eff * T * np.log(residual / (n_eff * T)) + _penalty(k)

        null_bic = _bic(0.0, 0)

        # --- Score all individual sources for multi-start seeds ---
        single_idx = np.arange(n_sources, dtype=np.intp).reshape(-1, 1)
        single_scores = _batch_score(Gram, Q, single_idx)
        seed_order = np.argsort(single_scores)[::-1]
        n_starts_eff = min(n_starts, n_sources)

        # --- Multi-start greedy forward selection ---
        global_best_bic = null_bic
        global_best_selected = None

        for start_i in range(n_starts_eff):
            seed = int(seed_order[start_i])
            selected = [seed]
            current_bic = _bic(single_scores[seed], 1)

            if current_bic >= null_bic:
                # This seed isn't even better than null model
                continue

            for k in range(2, min(k_max, n_sources) + 1):
                selected_set = set(selected)
                remaining = np.array(
                    [j for j in range(n_sources) if j not in selected_set],
                    dtype=np.intp,
                )
                if len(remaining) == 0:
                    break

                candidates = np.array(
                    [tuple(sorted(selected + [j])) for j in remaining],
                    dtype=np.intp,
                )
                scores = _batch_score(Gram, Q, candidates)

                best_idx = np.argmax(scores)
                new_bic = _bic(scores[best_idx], k)

                if new_bic < current_bic:
                    current_bic = new_bic
                    selected = list(candidates[best_idx])
                else:
                    break

            # --- Coordinate-wise refinement ---
            current_score = _score_single(Gram, Q, selected)

            for _ in range(n_refine_iters):
                improved = False
                for pos_idx in range(len(selected)):
                    base = [s for i, s in enumerate(selected) if i != pos_idx]
                    base_set = set(base)
                    remaining = np.array(
                        [j for j in range(n_sources) if j not in base_set],
                        dtype=np.intp,
                    )

                    candidates = np.array(
                        [tuple(sorted(base + [j])) for j in remaining],
                        dtype=np.intp,
                    )
                    scores = _batch_score(Gram, Q, candidates)

                    best_idx = np.argmax(scores)
                    if scores[best_idx] > current_score + 1e-10:
                        selected = list(candidates[best_idx])
                        current_score = scores[best_idx]
                        improved = True

                if not improved:
                    break

            # Recompute BIC after refinement (K might be same but score changed)
            final_bic = _bic(current_score, len(selected))
            if final_bic < global_best_bic:
                global_best_bic = final_bic
                global_best_selected = list(selected)

        if global_best_selected is None:
            # Fallback: best individual source
            global_best_selected = [int(seed_order[0])]

        selected_sources = np.array(global_best_selected, dtype=np.intp)

        # --- Construct ML inverse operator ---
        G_sel = G_w[:, selected_sources]
        inv_w = np.linalg.lstsq(G_sel, np.eye(n_eff), rcond=None)[0]

        inverse_operator_w = np.zeros((n_sources, n_eff))
        inverse_operator_w[selected_sources, :] = inv_w

        self.inverse_operators = [
            InverseOperator(inverse_operator_w @ wf.sensor_transform, self.name),
        ]
        return self



import numpy as np
from scipy.sparse import csr_matrix

from invert.solvers.bayesian.cmem import (
    _cmem,
    _compute_msp_coefficients,
    _data_driven_parcellation,
)


def test_cmem_mne_identity_small():
    """cMEM uses a Woodbury-form ridge/MNE solve; it should match the direct form."""
    rng = np.random.default_rng(0)
    m, n, t = 9, 25, 7
    L = rng.standard_normal((m, n))
    Y = rng.standard_normal((m, t))

    fro2 = float(np.sum(L * L))  # trace(L^T L)
    reg = 0.1 * fro2 / n

    direct = np.linalg.solve(L.T @ L + reg * np.eye(n), L.T @ Y)
    woodbury = L.T @ np.linalg.solve(L @ L.T + reg * np.eye(m), Y)

    np.testing.assert_allclose(direct, woodbury, rtol=1e-10, atol=1e-10)


def test_cmem_runs_small():
    rng = np.random.default_rng(0)
    m, n, t = 10, 40, 6
    L = rng.standard_normal((m, n)) * 0.1
    Y = rng.standard_normal((m, t))

    rows, cols = [], []
    for i in range(n):
        if i - 1 >= 0:
            rows.append(i)
            cols.append(i - 1)
        if i + 1 < n:
            rows.append(i)
            cols.append(i + 1)
    A = csr_matrix((np.ones(len(rows)), (rows, cols)), shape=(n, n))

    J, parcels = _cmem(Y, L, A=A, num_parcels=8, max_iter=3, batch_size=3)
    assert J.shape == (n, t)
    assert parcels.shape == (n,)
    assert np.isfinite(J).all()


def test_cmem_parcellation_respects_num_parcels_with_adjacency():
    rng = np.random.default_rng(0)
    m, n, t = 8, 500, 10
    L = rng.standard_normal((m, n)) * 0.1
    Y = rng.standard_normal((m, t))

    rows, cols = [], []
    for i in range(n):
        if i - 1 >= 0:
            rows.append(i)
            cols.append(i - 1)
        if i + 1 < n:
            rows.append(i)
            cols.append(i + 1)
    A = csr_matrix((np.ones(len(rows)), (rows, cols)), shape=(n, n))

    msp = _compute_msp_coefficients(Y, L)
    num_parcels = 40
    parcels = _data_driven_parcellation(
        Y, L, num_parcels=num_parcels, A=A, msp_scores=msp
    )
    assert np.unique(parcels).size == num_parcels


def test_cmem_does_not_collapse_under_noise():
    """Regression: Sigma calibration must not drive the solution to ~zero."""
    rng = np.random.default_rng(0)
    m, n, t = 16, 200, 30
    L = rng.standard_normal((m, n)) * 0.1

    active = int(rng.integers(0, n))
    S = np.zeros((n, t))
    S[active] = rng.standard_normal(t)

    Y_clean = L @ S
    noise_rel = 0.30
    Y = Y_clean + rng.standard_normal(Y_clean.shape) * (
        noise_rel * float(np.std(Y_clean))
    )

    rows, cols = [], []
    for i in range(n):
        if i - 1 >= 0:
            rows.append(i)
            cols.append(i - 1)
        if i + 1 < n:
            rows.append(i)
            cols.append(i + 1)
    A = csr_matrix((np.ones(len(rows)), (rows, cols)), shape=(n, n))

    J, _ = _cmem(Y, L, A=A, num_parcels=20, max_iter=15, batch_size=10)
    residual = Y - L @ J
    r2 = 1.0 - float(np.sum(residual**2)) / float(np.sum(Y**2))
    assert r2 > 0.9



"""3D-like glass brain visualization using Matplotlib with depth cues.

Simulates a 3D view of the brain by projecting sources to 2D and scaling
markers based on their depth (distance to viewer).
"""

from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import Normalize
from scipy.spatial import ConvexHull


def _get_source_positions(src):
    """Extract 3D positions for all source-space vertices."""
    positions = []
    for hemi in src:
        positions.append(hemi["rr"][hemi["vertno"]])
    return np.concatenate(positions, axis=0)


def _rotate_points(points, azimuth, elevation):
    """Rotate points by azimuth (around Z) and elevation (around X)."""
    # Convert to radians
    az = np.radians(azimuth)
    el = np.radians(elevation)

    # Rotation around Z axis (Azimuth)
    Rz = np.array(
        [[np.cos(az), -np.sin(az), 0], [np.sin(az), np.cos(az), 0], [0, 0, 1]]
    )

    # Rotation around X axis (Elevation)
    Rx = np.array(
        [[1, 0, 0], [0, np.cos(el), -np.sin(el)], [0, np.sin(el), np.cos(el)]]
    )

    # Apply rotations: R = Rx @ Rz
    # Points are (N, 3). We want (R @ P.T).T = P @ R.T
    R = Rx @ Rz
    return points @ R.T


def plot_3d_glass_brain(
    stc,
    src,
    time_idx=None,
    threshold=0.2,
    cmap="magma",
    alpha=0.9,
    marker_size=30,
    brain_alpha=0.05,
    brain_color="0.7",
    figsize=(8, 8),
    title=None,
    ax=None,
    view=(90, 10),  # Azimuth, Elevation
    depth_scale=True,
):
    """Plot a 3D-like glass brain visualization.

    Parameters
    ----------
    stc : mne.SourceEstimate
        Source estimate.
    src : mne.SourceSpaces
        Source space.
    time_idx : int | None
        Time index to plot.
    threshold : float
        Threshold fraction (0.0 to 1.0) of peak activation.
    cmap : str
        Colormap.
    alpha : float
        Opacity of active sources.
    marker_size : float
        Base marker size.
    brain_alpha : float
        Opacity of the inactive source cloud.
    brain_color : str
        Color of the brain outline/cloud.
    figsize : tuple
        Figure size.
    title : str | None
        Title.
    ax : matplotlib.axes.Axes | None
        Axes to plot on.
    view : tuple
        (azimuth, elevation) in degrees.
    depth_scale : bool
        Whether to scale marker size by depth.

    Returns
    -------
    fig : matplotlib.figure.Figure
        The figure.
    """
    all_pos = _get_source_positions(src)
    all_pos_mm = all_pos * 1e3

    # Rotate all points
    az, el = view
    rotated_pos = _rotate_points(all_pos_mm, az, el)

    # Extract data
    data = stc.data
    if data.ndim == 1:
        values = data
    else:
        if time_idx is None:
            time_idx = np.argmax(np.max(np.abs(data), axis=0))
        values = data[:, time_idx]
    values = np.abs(values)

    # Get active positions
    active_pos_list = []
    for i, hemi in enumerate(src):
        active_pos_list.append(hemi["rr"][stc.vertices[i]])
    active_pos = np.concatenate(active_pos_list, axis=0) * 1e3

    # Rotate active positions using same rotation
    active_pos_rot = _rotate_points(active_pos, az, el)

    # Threshold
    peak = values.max()
    if peak > 0:
        mask = values >= threshold * peak
    else:
        mask = np.ones(len(values), dtype=bool)

    active_pos_rot = active_pos_rot[mask]
    values = values[mask]

    # Sort by depth (Y coordinate after rotation? Or Z?
    # In matplotlib 2D, we plot X vs Y (or Z).
    # Let's project to X-Z plane (coronal-like) or X-Y (axial-like).
    # If we rotate such that the viewer is looking down the Y axis (or Z axis).
    # Let's assume we plot X (horizontal) and Z (vertical) of the rotated points.
    # The depth is Y (going into the screen).
    # Standard: X=Right, Y=Anterior, Z=Superior.
    # After rotation, we plot X' and Z'. Y' is depth.
    # If Y' is depth, larger Y' means closer or further depending on convention.
    # Let's assume positive Y is "into screen" or "away".
    # Actually, in right-handed system:
    # If we look from +Y towards origin, then X is right, Z is up.
    # So Y is depth. Larger Y = closer to viewer if we look from +inf Y.

    # Let's define projection:
    x_proj = rotated_pos[:, 0]
    y_proj = rotated_pos[:, 2]  # Plot Z on vertical axis
    depth = rotated_pos[:, 1]  # Y is depth

    # Active sources projection
    act_x = active_pos_rot[:, 0]
    act_y = active_pos_rot[:, 2]
    act_depth = active_pos_rot[:, 1]

    # Sort active sources by depth (furthest first -> smallest Y first if looking from +Y)
    # If looking from +Y, larger Y is closer. We want to plot furthest (small Y) first.
    sort_idx = np.argsort(act_depth)
    act_x = act_x[sort_idx]
    act_y = act_y[sort_idx]
    act_depth = act_depth[sort_idx]
    values = values[sort_idx]

    # Setup plot
    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)
    else:
        fig = ax.figure

    # 1. Draw brain outline (Convex Hull of projected points)
    points_2d = np.column_stack((x_proj, y_proj))
    try:
        hull = ConvexHull(points_2d)
        for simplex in hull.simplices:
            ax.plot(
                points_2d[simplex, 0],
                points_2d[simplex, 1],
                color=brain_color,
                alpha=0.5,
                linewidth=1.5,
            )
    except Exception:
        pass

    # 2. Draw faint cloud of all sources
    ax.scatter(
        x_proj,
        y_proj,
        c=brain_color,
        s=1,
        alpha=brain_alpha,
        edgecolors="none",
        zorder=1,
    )

    # 3. Draw active sources
    if len(values) > 0:
        norm = Normalize(vmin=0, vmax=peak if peak > 0 else 1)

        sizes = marker_size
        if depth_scale:
            # Scale size by depth.
            # Normalize depth to [0, 1] or similar.
            # Larger depth (closer) -> larger size.
            # act_depth range:
            d_min, d_max = depth.min(), depth.max()
            if d_max > d_min:
                d_norm = (act_depth - d_min) / (d_max - d_min)  # 0 to 1
                # Scale factor: e.g. 0.5 to 2.0
                scale = 0.5 + 2.0 * d_norm
                sizes = marker_size * (scale**2)  # Area scales with square

        ax.scatter(
            act_x,
            act_y,
            c=values,
            cmap=cmap,
            norm=norm,
            s=sizes,
            alpha=alpha,
            edgecolors="none",
            zorder=10,
        )

        # Add colorbar if this is the only axis or requested
        # (Handling colorbar is tricky with subplots, usually done outside)

    # Styling
    ax.set_aspect("equal")
    ax.axis("off")
    if title:
        ax.set_title(title, fontweight="bold")

    return fig



from __future__ import annotations

import importlib.util as _ilu
import json
import logging
import os
import random
import time
from concurrent.futures import ThreadPoolExecutor, as_completed
from datetime import datetime
from pathlib import Path
from typing import Any

import mne
import numpy as np
from pydantic import BaseModel
from tqdm import tqdm

_spec = _ilu.spec_from_file_location(
    "invert.evaluate.evaluate",
    str(Path(__file__).resolve().parents[1] / "evaluate" / "evaluate.py"),
)
assert _spec is not None
assert _spec.loader is not None
_eval_mod = _ilu.module_from_spec(_spec)
_spec.loader.exec_module(_eval_mod)
evaluate_all = _eval_mod.evaluate_all
from invert.simulate import SimulationConfig, SimulationGenerator  # noqa: E402
from invert.solver_registry import (  # noqa: E402
    get_solver_class,
    get_solver_spec,
    iter_solver_specs,
)
from invert.solvers.base import BaseSolver  # noqa: E402
from invert.util.util import pos_from_forward  # noqa: E402

from .datasets import BENCHMARK_DATASETS, DatasetConfig  # noqa: E402
from .depth_tuning import (  # noqa: E402
    RECOMMENDED_DEPTH_BY_CATEGORY,
    RECOMMENDED_DEPTH_BY_SOLVER,
    build_depth_solver_params,
)

logger = logging.getLogger(__name__)


def _expects_simulation_config(solver_cls: type[BaseSolver]) -> bool:
    """Return True if solver.make_inverse_operator expects a SimulationConfig.

    Neural-network solvers in this codebase train from simulated data and expose
    a signature like: make_inverse_operator(forward, simulation_config, ...).
    """
    import inspect

    try:
        sig = inspect.signature(solver_cls.make_inverse_operator)
    except (TypeError, ValueError):
        return False

    params = list(sig.parameters.values())
    # params[0] is "self"; params[1] is typically "forward"
    for p in params[2:]:
        if p.kind in (inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD):
            continue
        return p.name == "simulation_config"
    return False


def _default_nn_batch_size(forward: mne.Forward) -> int:
    """Default ANN training batch size: ~12x number of dipoles (min 4096).

    Larger batches provide more diverse training data per epoch, which
    significantly improves ANN solver quality (see ann_testbench round-2/3).
    """
    try:
        n_dipoles = int(forward["sol"]["data"].shape[1])
    except Exception:
        return 1
    return max(4096, 12 * n_dipoles)


def _generate_solver_categories() -> dict[str, list[str]]:
    """Automatically generate solver categories from solver_registry.

    Categories are determined by the solver's module path:
    - beamformers/ -> beamformer
    - bayesian/ -> bayesian
    - minimum_norm/ -> loreta (if solver_id contains "loreta") or minimum_norm
    - music/ -> music
    - matching_pursuit/ -> matching_pursuit
    - hybrids/ -> other
    - neural_networks/ -> neural_networks
    - dipoles/ -> other
    - random_noise -> other
    """
    categories: dict[str, list[str]] = {}

    for spec in iter_solver_specs(include_internal=False, include_unavailable=True):
        module_path = spec.module_path
        solver_id = spec.solver_id

        # Extract directory from module path
        # e.g., "invert.solvers.beamformers.esmv" -> "beamformers"
        parts = module_path.split(".")
        if len(parts) < 3 or parts[0] != "invert" or parts[1] != "solvers":
            categories.setdefault("other", []).append(solver_id)
            continue

        directory = parts[2]

        if directory == "beamformers":
            category = "beamformer"
        elif directory == "bayesian":
            category = "bayesian"
        elif directory == "minimum_norm":
            if "loreta" in solver_id:
                category = "loreta"
            else:
                category = "minimum_norm"
        elif directory == "music":
            category = "music"
        elif directory == "matching_pursuit":
            category = "matching_pursuit"
        elif directory == "neural_networks":
            category = "neural_networks"
        else:
            category = "other"

        categories.setdefault(category, []).append(solver_id)

    for category in categories:
        categories[category].sort()

    return categories


SOLVER_CATEGORIES = _generate_solver_categories()


def get_solver_category(solver_name: str) -> str:
    """Get the category for a solver (accepts any alias).

    Parameters
    ----------
    solver_name : str
        The solver's canonical id or any known alias.

    Returns
    -------
    str
        The category name, or "other" if not found in any category.
    """
    try:
        sid = get_solver_spec(solver_name).solver_id
    except ValueError:
        return "other"
    for category, solvers in SOLVER_CATEGORIES.items():
        if sid in solvers:
            return category
    return "other"


def resolve_solvers(
    solvers: list[str] | None = None,
    categories: list[str] | None = None,
    exclude: list[str] | None = None,
) -> list[str]:
    """Resolve a list of solver names from explicit names and/or categories.

    Parameters
    ----------
    solvers : list of str, optional
        Explicit solver names or aliases (e.g. ``["MNE", "Champagne"]``).
    categories : list of str, optional
        Category names to include (e.g. ``["beamformer", "loreta"]``).
        Use ``"all"`` to include every registered (non-internal) solver.
    exclude : list of str, optional
        Solver names or aliases to exclude from the result.

    Returns
    -------
    list of str
        Deduplicated, order-preserved list of canonical solver_ids.
    """
    result: list[str] = []
    seen: set[str] = set()

    exclude_ids: set[str] = set()
    if exclude:
        for name in exclude:
            exclude_ids.add(get_solver_spec(name).solver_id)

    if categories:
        for cat in categories:
            if cat == "all":
                names = [
                    spec.solver_id
                    for spec in iter_solver_specs(
                        include_internal=False, include_unavailable=True
                    )
                ]
            elif cat in SOLVER_CATEGORIES:
                names = SOLVER_CATEGORIES[cat]
            else:
                raise ValueError(
                    f"Unknown category {cat!r}. "
                    f"Available: {sorted(SOLVER_CATEGORIES)} or 'all'"
                )
            for n in names:
                if n not in seen and n not in exclude_ids:
                    seen.add(n)
                    result.append(n)

    if solvers:
        for name in solvers:
            sid = get_solver_spec(name).solver_id
            if sid not in seen and sid not in exclude_ids:
                seen.add(sid)
                result.append(sid)

    return result


# Default solvers when nothing is specified
_DEFAULT_SOLVERS = ["mne", "eloreta", "lcmv", "champagne-mackay", "random-noise"]


# ---------------------------------------------------------------------------
# Worker functions for parallelization (must be at module level for pickling)
# ---------------------------------------------------------------------------


def _apply_inverse_worker(
    idx: int,
    inv_op_matrix: np.ndarray,
    x_sample: np.ndarray,
    y_sample: np.ndarray,
    adjacency,
    pos: np.ndarray,
) -> tuple[int, dict]:
    """Worker for applying precomputed inverse operator.

    Returns (sample_idx, metrics_dict) for ordering results.
    """
    y_pred = inv_op_matrix @ x_sample
    metrics = evaluate_all(y_sample, y_pred, adjacency, adjacency, pos, pos)
    return idx, metrics


def _split_solver_params(
    solver_params: dict[str, Any] | None,
) -> tuple[dict[str, Any], dict[str, Any], dict[str, Any]]:
    """Split params into init/make/apply dictionaries using prefixes."""
    init_params: dict[str, Any] = {}
    make_params: dict[str, Any] = {}
    apply_params: dict[str, Any] = {}
    if not solver_params:
        return init_params, make_params, apply_params
    for key, value in solver_params.items():
        if key.startswith("init__"):
            init_params[key.removeprefix("init__")] = value
        elif key.startswith("apply__"):
            apply_params[key.removeprefix("apply__")] = value
        elif key.startswith("make__"):
            make_params[key.removeprefix("make__")] = value
        else:
            make_params[key] = value
    return init_params, make_params, apply_params


def _compute_and_apply_worker(
    idx: int,
    solver_module: str,
    solver_class: str,
    forward: mne.Forward,
    info: mne.Info,
    x_sample: np.ndarray,
    y_sample: np.ndarray,
    adjacency,
    pos: np.ndarray,
    require_data: bool,
    noise_cov: mne.Covariance | None = None,
    solver_params: dict[str, Any] | None = None,
) -> tuple[int, dict]:
    """Worker for computing fresh inverse operator per sample.

    Returns (sample_idx, metrics_dict) for ordering results.
    """
    import importlib

    mod = importlib.import_module(solver_module)
    solver_cls = getattr(mod, solver_class)
    init_params, make_params, apply_params = _split_solver_params(solver_params)
    solver = solver_cls(**init_params)

    evoked = mne.EvokedArray(x_sample, info, tmin=0.0, verbose=0)
    _make_inverse_operator_with_covariance(
        solver=solver,
        forward=forward,
        require_data=require_data,
        evoked=evoked,
        alpha="auto",
        noise_cov=noise_cov,
        make_kwargs=make_params,
    )
    stc = solver.apply_inverse_operator(evoked, **apply_params)
    y_pred = stc.data
    metrics = evaluate_all(y_sample, y_pred, adjacency, adjacency, pos, pos)
    return idx, metrics


def _make_mne_covariance(
    noise_cov: np.ndarray,
    info: mne.Info,
    *,
    nfree: int = 1,
) -> mne.Covariance:
    """Build an MNE covariance with channel metadata from info."""
    cov = np.asarray(noise_cov, dtype=float)
    ch_names = list(info["ch_names"])
    if cov.ndim != 2 or cov.shape[0] != cov.shape[1]:
        msg = f"noise_cov must be square, got shape {cov.shape}"
        raise ValueError(msg)
    if cov.shape[0] != len(ch_names):
        msg = f"noise_cov has shape {cov.shape}, but info has {len(ch_names)} channels"
        raise ValueError(msg)
    return mne.Covariance(
        data=cov,
        names=ch_names,
        bads=list(info.get("bads", [])),
        projs=list(info.get("projs", [])),
        nfree=int(max(nfree, 1)),
    )


def _make_inverse_operator_with_covariance(
    *,
    solver: BaseSolver,
    forward: mne.Forward,
    require_data: bool,
    evoked: mne.Evoked | None,
    alpha: str | float,
    noise_cov: mne.Covariance | None,
    make_kwargs: dict[str, Any] | None = None,
) -> None:
    """Compute an inverse operator using an MNE covariance object."""
    kwargs = dict(make_kwargs or {})
    if require_data:
        assert evoked is not None
        solver.make_inverse_operator(
            forward, evoked, alpha=alpha, noise_cov=noise_cov, **kwargs
        )
    else:
        solver.make_inverse_operator(
            forward, alpha=alpha, noise_cov=noise_cov, **kwargs
        )


class TimingInfo(BaseModel):
    total_seconds: float
    make_inverse_operator_seconds: float | None = None
    apply_inverse_seconds: float | None = None


class SampleMetrics(BaseModel):
    mle: float
    emd: float
    sd: float
    ap: float
    correlation: float


class AggregateStats(BaseModel):
    mean: float
    std: float
    median: float
    # Optional for compact dashboard exports.
    worst_10_pct: float | None = None  # 10th/90th percentile of worst predictions


class BenchmarkResult(BaseModel):
    solver_name: str
    dataset_name: str
    category: str | None = None
    metrics: dict[str, AggregateStats]
    samples: list[SampleMetrics]
    timing: TimingInfo | None = None


class BenchmarkRunner:
    def __init__(
        self,
        forward: mne.Forward,
        info: mne.Info,
        solvers: list[str] | None = None,
        categories: list[str] | None = None,
        exclude_solvers: list[str] | None = None,
        datasets: dict[str, DatasetConfig] | None = None,
        n_samples: int = 50,
        n_jobs: int | None = None,
        random_seed: int | None = None,
        solver_params: dict[str, dict[str, Any]] | None = None,
        depth_by_category: dict[str, float] | None = None,
        depth_by_solver: dict[str, float] | None = None,
    ):
        self.forward = forward
        self.info = info
        if solvers is None and categories is None:
            exclude_ids: set[str] = set()
            if exclude_solvers:
                for name in exclude_solvers:
                    exclude_ids.add(get_solver_spec(name).solver_id)
            self.solvers = [s for s in _DEFAULT_SOLVERS if s not in exclude_ids]
        else:
            self.solvers = resolve_solvers(
                solvers=solvers, categories=categories, exclude=exclude_solvers
            )
        self.datasets = datasets or dict(BENCHMARK_DATASETS)
        self.n_samples = n_samples
        self.random_seed = random_seed
        # Normalize solver_params keys to canonical solver_ids.
        explicit_solver_params = {
            get_solver_spec(k).solver_id: v for k, v in (solver_params or {}).items()
        }
        # Apply package depth defaults unless explicitly disabled via {}.
        resolved_depth_by_category = (
            dict(RECOMMENDED_DEPTH_BY_CATEGORY)
            if depth_by_category is None
            else dict(depth_by_category)
        )
        resolved_depth_by_solver = (
            dict(RECOMMENDED_DEPTH_BY_SOLVER)
            if depth_by_solver is None
            else dict(depth_by_solver)
        )
        self.solver_params = build_depth_solver_params(
            solvers=self.solvers,
            depth_by_category=resolved_depth_by_category,
            depth_by_solver=resolved_depth_by_solver,
            base_solver_params=explicit_solver_params,
        )

        # Auto-detect n_jobs if not specified
        if n_jobs is None:
            cpu_count = os.cpu_count()
            self.n_jobs = max(1, cpu_count - 1) if cpu_count is not None else 1
        elif n_jobs == -1:
            self.n_jobs = os.cpu_count() or 1
        else:
            self.n_jobs = max(1, n_jobs)

        self._results: list[BenchmarkResult] = []

    def run(self) -> list[BenchmarkResult]:
        pos = pos_from_forward(self.forward)
        adjacency = mne.spatial_src_adjacency(self.forward["src"], verbose=0)
        results = []

        # Total number of (dataset, solver) combinations for overall progress
        total_combinations = len(self.datasets) * len(self.solvers)

        with tqdm(
            total=total_combinations, desc="Overall Progress", position=0
        ) as pbar_overall:
            for ds_name, ds_config in self.datasets.items():
                logger.info("Dataset: %s", ds_name)

                # Generate all samples for this dataset once
                sim_config = SimulationConfig(
                    batch_size=self.n_samples,
                    n_sources=ds_config.n_sources,
                    n_orders=ds_config.n_orders,
                    snr_range=ds_config.snr_range,
                    n_timepoints=ds_config.n_timepoints,
                    source_spatial_model=ds_config.source_spatial_model,
                    patch_rank=ds_config.patch_rank,
                    random_seed=self.random_seed,
                    estimate_noise_cov=True,
                    return_noise_cov=True,
                )
                gen = SimulationGenerator(self.forward, config=sim_config)
                x_batch, y_batch, sim_info = next(gen.generate())
                # Per-sample noise covariance (scaled to match the actual
                # noise level in each sample).  Solvers that recompute per
                # sample receive the matching covariance; non-recompute
                # solvers receive sample-0's estimate (consistent with
                # already using x_batch[0] for their inverse operator).
                noise_covs = [
                    _make_mne_covariance(
                        sim_info.iloc[j]["noise_cov_est_scaled"],
                        self.info,
                        nfree=1,
                    )
                    for j in range(len(sim_info))
                ]
                noise_cov = noise_covs[0]

                for solver_name in self.solvers:
                    logger.info("  Solver: %s", solver_name)
                    spec = get_solver_spec(solver_name)
                    solver_cls = get_solver_class(solver_name)
                    init_params, make_params, apply_params = _split_solver_params(
                        self.solver_params.get(solver_name, {})
                    )
                    merged_init = {**spec.default_kwargs, **init_params}
                    solver = solver_cls(**merged_init)

                    # Best-effort determinism for fair comparisons when a seed is provided.
                    if self.random_seed is not None:
                        seed = int(self.random_seed)
                        random.seed(seed)
                        np.random.seed(seed)
                        try:  # torch is optional
                            import torch  # type: ignore[import-not-found]

                            torch.manual_seed(seed)
                            if torch.cuda.is_available():
                                torch.cuda.manual_seed_all(seed)
                        except Exception:
                            pass

                    # Neural-network solvers train from SimulationConfig and then
                    # apply the trained model to each sample.
                    t_solver_start = time.perf_counter()
                    timing_info: TimingInfo | None = None

                    if _expects_simulation_config(solver_cls):
                        train_sim_config = sim_config.model_copy(
                            update={"batch_size": _default_nn_batch_size(self.forward)}
                        )
                        logger.info(
                            "ANN training batch_size=%d for %s",
                            int(train_sim_config.batch_size),
                            solver_name,
                        )
                        params = dict(make_params)
                        alpha = params.pop("alpha", "auto")
                        t_make = time.perf_counter()
                        solver.make_inverse_operator(
                            self.forward, train_sim_config, alpha=alpha, **params
                        )
                        make_seconds = time.perf_counter() - t_make
                        t_apply = time.perf_counter()
                        sample_metrics: list[SampleMetrics] = []
                        for i in tqdm(
                            range(self.n_samples),
                            desc=f"{ds_name}/{solver_name}",
                            position=1,
                            leave=False,
                        ):
                            evoked = mne.EvokedArray(
                                x_batch[i], self.info, tmin=0.0, verbose=0
                            )
                            stc = solver.apply_inverse_operator(evoked, **apply_params)
                            y_pred = stc.data
                            metrics = evaluate_all(
                                y_batch[i], y_pred, adjacency, adjacency, pos, pos
                            )
                            sample_metrics.append(self._metrics_from_dict(metrics))
                        apply_seconds = time.perf_counter() - t_apply
                        total_seconds = time.perf_counter() - t_solver_start
                        timing_info = TimingInfo(
                            total_seconds=total_seconds,
                            make_inverse_operator_seconds=make_seconds,
                            apply_inverse_seconds=apply_seconds,
                        )

                    # Parallelize based on require_recompute
                    elif not solver.require_recompute:
                        # Compute inverse operator once, then parallelize application
                        evoked = None
                        if solver.require_data:
                            evoked = mne.EvokedArray(
                                x_batch[0], self.info, tmin=0.0, verbose=0
                            )
                        t_make = time.perf_counter()
                        _make_inverse_operator_with_covariance(
                            solver=solver,
                            forward=self.forward,
                            require_data=bool(solver.require_data),
                            evoked=evoked,
                            alpha="auto",
                            noise_cov=noise_cov,
                            make_kwargs=make_params,
                        )
                        make_seconds = time.perf_counter() - t_make

                        # Some solvers do not expose inverse operators, or may leave the
                        # list empty for data-dependent paths. In both cases, fall back
                        # to direct per-sample application.
                        t_apply = time.perf_counter()
                        inverse_ops = getattr(solver, "inverse_operators", None)
                        if not inverse_ops:
                            # Fall back to direct application for each sample
                            sample_metrics = []
                            for i in range(len(x_batch)):
                                evoked = mne.EvokedArray(
                                    x_batch[i], self.info, tmin=0.0, verbose=0
                                )
                                stc = solver.apply_inverse_operator(
                                    evoked, **apply_params
                                )
                                y_pred = stc.data
                                metrics = evaluate_all(
                                    y_batch[i], y_pred, adjacency, adjacency, pos, pos
                                )
                                sample_metrics.append(self._metrics_from_dict(metrics))
                        else:
                            # Select optimal regularization via L-curve/GCV
                            n_ops = len(inverse_ops)
                            if get_solver_category(solver_name) == "beamformer":
                                # Beamformer regularization selection: do not use
                                # residual-based GCV/L-curve criteria.
                                optimal_idx = 0
                            elif n_ops > 1:
                                try:
                                    # Aggregate GCV scores across multiple samples
                                    # for more robust regularization selection.
                                    n_gcv = min(5, len(x_batch))
                                    all_scores = np.stack(
                                        [
                                            solver.compute_gcv_scores(x_batch[k])
                                            for k in range(n_gcv)
                                        ]
                                    )
                                    median_scores = np.median(all_scores, axis=0)
                                    valid = np.isfinite(median_scores)
                                    if np.any(valid):
                                        valid_idx = np.where(valid)[0]
                                        optimal_idx = valid_idx[
                                            np.argmin(median_scores[valid])
                                        ]
                                    else:
                                        optimal_idx = len(median_scores) // 2
                                except Exception as exc:
                                    logger.warning(
                                        "Regularisation selection failed for %s on %s; "
                                        "falling back to first inverse operator (%s).",
                                        solver_name,
                                        ds_name,
                                        exc,
                                    )
                                    optimal_idx = 0
                            else:
                                optimal_idx = 0
                            if optimal_idx < 0 or optimal_idx >= n_ops:
                                clipped_idx = int(np.clip(optimal_idx, 0, n_ops - 1))
                                logger.warning(
                                    "Regularisation index %d out of bounds for %s on %s "
                                    "(n_ops=%d); using idx=%d instead.",
                                    optimal_idx,
                                    solver_name,
                                    ds_name,
                                    n_ops,
                                    clipped_idx,
                                )
                                optimal_idx = clipped_idx
                            inv_op = inverse_ops[optimal_idx]

                            # Extract the inverse operator matrix (numpy array)
                            inv_op_matrix = inv_op.data[0]

                            # Parallel application
                            sample_metrics = self._run_parallel_apply(
                                inv_op_matrix,
                                x_batch,
                                y_batch,
                                adjacency,
                                pos,
                                ds_name,
                                solver_name,
                            )
                        apply_seconds = time.perf_counter() - t_apply
                        total_seconds = time.perf_counter() - t_solver_start
                        timing_info = TimingInfo(
                            total_seconds=total_seconds,
                            make_inverse_operator_seconds=make_seconds,
                            apply_inverse_seconds=apply_seconds,
                        )
                    else:
                        # Parallelize full computation (require_recompute=True)
                        # Merge default_kwargs as init__ prefixed keys so the
                        # worker receives them through _split_solver_params.
                        merged_params = {
                            f"init__{k}": v for k, v in spec.default_kwargs.items()
                        }
                        merged_params.update(self.solver_params.get(solver_name, {}))

                        sample_metrics = self._run_parallel_compute(
                            spec.module_path,
                            spec.class_name,
                            self.forward,
                            self.info,
                            x_batch,
                            y_batch,
                            adjacency,
                            pos,
                            solver.require_data,
                            ds_name,
                            solver_name,
                            noise_covs,
                            merged_params,
                        )
                        total_seconds = time.perf_counter() - t_solver_start
                        timing_info = TimingInfo(total_seconds=total_seconds)

                    result = self._aggregate(solver_name, ds_name, sample_metrics)
                    result.timing = timing_info
                    results.append(result)
                    pbar_overall.update(1)
                    pbar_overall.set_postfix(
                        {"dataset": ds_name, "solver": solver_name}
                    )

        self._results = results
        return results

    def _run_parallel_apply(
        self,
        inv_op_matrix: np.ndarray,
        x_batch: np.ndarray,
        y_batch: np.ndarray,
        adjacency,
        pos: np.ndarray,
        ds_name: str,
        solver_name: str,
    ) -> list[SampleMetrics]:
        """Parallelize inverse operator application (require_recompute=False)."""
        if self.n_jobs == 1:
            sample_metrics = []
            for i in tqdm(
                range(self.n_samples),
                desc=f"{ds_name}/{solver_name}",
                position=1,
                leave=False,
            ):
                _, metrics = _apply_inverse_worker(
                    i, inv_op_matrix, x_batch[i], y_batch[i], adjacency, pos
                )
                sample_metrics.append(self._metrics_from_dict(metrics))
            return sample_metrics

        sample_metrics_dict = {}
        with ThreadPoolExecutor(max_workers=self.n_jobs) as executor:
            futures = {
                executor.submit(
                    _apply_inverse_worker,
                    i,
                    inv_op_matrix,
                    x_batch[i],
                    y_batch[i],
                    adjacency,
                    pos,
                ): i
                for i in range(self.n_samples)
            }

            with tqdm(
                total=self.n_samples,
                desc=f"{ds_name}/{solver_name}",
                position=1,
                leave=False,
            ) as pbar:
                for future in as_completed(futures):
                    try:
                        idx, metrics = future.result()
                        sample_metrics_dict[idx] = self._metrics_from_dict(metrics)
                    except Exception as e:
                        logger.error(f"Sample {futures[future]} failed: {e}")
                        idx = futures[future]
                        sample_metrics_dict[idx] = SampleMetrics(
                            mle=float("nan"),
                            emd=float("nan"),
                            sd=float("nan"),
                            ap=float("nan"),
                            correlation=float("nan"),
                        )
                    pbar.update(1)

        return [sample_metrics_dict[i] for i in range(self.n_samples)]

    def _run_parallel_compute(
        self,
        solver_module: str,
        solver_class: str,
        forward: mne.Forward,
        info: mne.Info,
        x_batch: np.ndarray,
        y_batch: np.ndarray,
        adjacency,
        pos: np.ndarray,
        require_data: bool,
        ds_name: str,
        solver_name: str,
        noise_covs: list[mne.Covariance] | None = None,
        solver_params: dict[str, Any] | None = None,
    ) -> list[SampleMetrics]:
        """Parallelize full computation (require_recompute=True)."""
        if self.n_jobs == 1:
            sample_metrics = []
            for i in tqdm(
                range(self.n_samples),
                desc=f"{ds_name}/{solver_name}",
                position=1,
                leave=False,
            ):
                nc = noise_covs[i] if noise_covs else None
                _, metrics = _compute_and_apply_worker(
                    i,
                    solver_module,
                    solver_class,
                    forward,
                    info,
                    x_batch[i],
                    y_batch[i],
                    adjacency,
                    pos,
                    require_data,
                    nc,
                    solver_params,
                )
                sample_metrics.append(self._metrics_from_dict(metrics))
            return sample_metrics

        sample_metrics_dict = {}
        with ThreadPoolExecutor(max_workers=self.n_jobs) as executor:
            futures = {
                executor.submit(
                    _compute_and_apply_worker,
                    i,
                    solver_module,
                    solver_class,
                    forward,
                    info,
                    x_batch[i],
                    y_batch[i],
                    adjacency,
                    pos,
                    require_data,
                    noise_covs[i] if noise_covs else None,
                    solver_params,
                ): i
                for i in range(self.n_samples)
            }

            with tqdm(
                total=self.n_samples,
                desc=f"{ds_name}/{solver_name}",
                position=1,
                leave=False,
            ) as pbar:
                for future in as_completed(futures):
                    try:
                        idx, metrics = future.result()
                        sample_metrics_dict[idx] = self._metrics_from_dict(metrics)
                    except Exception as e:
                        logger.error(f"Sample {futures[future]} failed: {e}")
                        idx = futures[future]
                        sample_metrics_dict[idx] = SampleMetrics(
                            mle=float("nan"),
                            emd=float("nan"),
                            sd=float("nan"),
                            ap=float("nan"),
                            correlation=float("nan"),
                        )
                    pbar.update(1)

        return [sample_metrics_dict[i] for i in range(self.n_samples)]

    @staticmethod
    def _metrics_from_dict(m: dict) -> SampleMetrics:
        return SampleMetrics(
            mle=float(m["Mean_Localization_Error"]),
            emd=float(m["EMD"]),
            sd=float(m["sd"]),
            ap=float(m["average_precision"]),
            correlation=float(m["correlation"]),
        )

    @staticmethod
    def _aggregate(
        solver_name: str,
        dataset_name: str,
        samples: list[SampleMetrics],
    ) -> BenchmarkResult:
        # Metrics where higher is better (worst = 10th percentile)
        # Others are lower is better (worst = 90th percentile)
        higher_is_better = {"average_precision", "correlation"}

        arrays = {
            "mean_localization_error": np.array([s.mle for s in samples]),
            "emd": np.array([s.emd for s in samples]),
            "spatial_dispersion": np.array([s.sd for s in samples]),
            "average_precision": np.array([s.ap for s in samples]),
            "correlation": np.array([s.correlation for s in samples]),
        }
        metrics = {}
        for key, arr in arrays.items():
            # For higher-is-better metrics, worst 10% = 10th percentile (lowest)
            # For lower-is-better metrics, worst 10% = 90th percentile (highest)
            worst_pct = 10 if key in higher_is_better else 90
            metrics[key] = AggregateStats(
                mean=float(np.nanmean(arr)),
                std=float(np.nanstd(arr)),
                median=float(np.nanmedian(arr)),
                worst_10_pct=float(np.nanpercentile(arr, worst_pct)),
            )
        return BenchmarkResult(
            solver_name=solver_name,
            dataset_name=dataset_name,
            category=get_solver_category(solver_name),
            metrics=metrics,
            samples=samples,
        )

    def _compute_best_solvers(self) -> dict[str, Any]:
        """Compute best solver per dataset for each metric.

        Returns
        -------
        dict
            Structure: {dataset_name: {metric_name: {"solver": solver_name, "value": metric_value}}}
        """
        # Metrics where lower is better
        lower_is_better = {"mean_localization_error", "emd", "spatial_dispersion"}
        # Metrics where higher is better
        higher_is_better = {"average_precision", "correlation"}

        datasets = sorted(set(r.dataset_name for r in self._results))
        all_metrics: set[str] = set()
        for r in self._results:
            all_metrics.update(r.metrics.keys())

        best_solvers: dict[str, Any] = {}
        for dataset in datasets:
            best_solvers[dataset] = {}
            for metric in all_metrics:
                # Find all results for this dataset
                dataset_results = [
                    r for r in self._results if r.dataset_name == dataset
                ]

                if not dataset_results:
                    continue

                # Get metric values for all solvers on this dataset
                # Filter out NaN values
                solver_values = {}
                for r in dataset_results:
                    if metric in r.metrics:
                        value = r.metrics[metric].mean
                        if not np.isnan(value):
                            solver_values[r.solver_name] = value

                if not solver_values:
                    continue

                # Determine best based on metric type
                if metric in lower_is_better:
                    best_solver = min(solver_values.items(), key=lambda x: x[1])
                elif metric in higher_is_better:
                    best_solver = max(solver_values.items(), key=lambda x: x[1])
                else:
                    # Default to lower is better if unknown
                    best_solver = min(solver_values.items(), key=lambda x: x[1])

                best_solvers[dataset][metric] = {
                    "solver": best_solver[0],
                    "value": round(best_solver[1], 4),
                }

        return best_solvers

    def _compute_average_ranks(
        self,
    ) -> tuple[dict[str, dict[str, float]], dict[str, float]]:
        """Compute per-dataset and global average ranks for each solver.

        For each (dataset, metric) pair, solvers are ranked 1..N (best=1).
        Per-dataset rank: average rank across metrics for that dataset.
        Global rank: average of the per-dataset ranks.

        Returns
        -------
        tuple
            (per_dataset_ranks, global_ranks) where:
            - per_dataset_ranks: {dataset_name: {solver_name: avg_rank}}
            - global_ranks: {solver_name: avg_rank}
        """
        lower_is_better = {"mean_localization_error", "emd", "spatial_dispersion"}

        datasets = sorted(set(r.dataset_name for r in self._results))
        all_metrics: set[str] = set()
        for r in self._results:
            all_metrics.update(r.metrics.keys())

        # Collect ranks per dataset: dataset -> solver_name -> list of ranks
        dataset_solver_ranks: dict[str, dict[str, list[float]]] = {}

        for dataset in datasets:
            dataset_results = [r for r in self._results if r.dataset_name == dataset]
            dataset_solver_ranks[dataset] = {}
            for metric in all_metrics:
                # Gather (solver, value) pairs, skip NaN
                solver_values = []
                for r in dataset_results:
                    if metric in r.metrics:
                        val = r.metrics[metric].mean
                        if not np.isnan(val):
                            solver_values.append((r.solver_name, val))

                if not solver_values:
                    continue

                # Sort: ascending for lower-is-better, descending for higher-is-better
                reverse = metric not in lower_is_better
                solver_values.sort(key=lambda x: x[1], reverse=reverse)

                # Dense ranking: tied values get the same rank
                rank = 1
                for i, (solver_name, val) in enumerate(solver_values):
                    if i > 0 and val != solver_values[i - 1][1]:
                        rank = i + 1
                    dataset_solver_ranks[dataset].setdefault(solver_name, []).append(
                        rank
                    )

        # Per-dataset ranks: average across metrics for each dataset
        per_dataset_ranks: dict[str, dict[str, float]] = {}
        for dataset in datasets:
            per_dataset_ranks[dataset] = {
                name: round(float(np.mean(ranks)), 2)
                for name, ranks in sorted(
                    dataset_solver_ranks[dataset].items(),
                    key=lambda x: np.mean(x[1]),
                )
            }

        # Global ranks: average of per-dataset ranks
        solver_dataset_ranks: dict[str, list[float]] = {}
        for _dataset, solver_ranks in per_dataset_ranks.items():
            for solver_name, avg_rank in solver_ranks.items():
                solver_dataset_ranks.setdefault(solver_name, []).append(avg_rank)

        global_ranks = {
            name: round(float(np.mean(ranks)), 2)
            for name, ranks in sorted(
                solver_dataset_ranks.items(), key=lambda x: np.mean(x[1])
            )
        }

        return per_dataset_ranks, global_ranks

    def save(
        self,
        path: str | Path,
        *,
        compact: bool = False,
        name: str | None = None,
        description: str | None = None,
    ) -> None:
        path = Path(path)

        per_dataset_ranks, global_ranks = self._compute_average_ranks()

        datasets_payload = {
            key: cfg.model_dump() if isinstance(cfg, BaseModel) else dict(cfg)  # type: ignore[arg-type]
            for key, cfg in self.datasets.items()
        }

        m_electrodes: int | None = None
        n_leadfield_columns: int | None = None
        n_sources_space: int | None = None
        n_orient: int | None = None
        try:
            lf = self.forward["sol"]["data"]
            m_electrodes = int(lf.shape[0])
            n_leadfield_columns = int(lf.shape[1])
        except Exception:
            m_electrodes = None
            n_leadfield_columns = None

        try:
            raw_nsource = self.forward.get("nsource")  # type: ignore[call-arg]
            if raw_nsource is not None:
                n_sources_space = int(raw_nsource)
        except Exception:
            n_sources_space = None

        if n_sources_space is None:
            try:
                src = self.forward.get("src")  # type: ignore[call-arg]
                if isinstance(src, (list, tuple)):
                    n_sources_space = int(
                        sum(
                            len(s.get("vertno", [])) for s in src if isinstance(s, dict)
                        )
                    )
            except Exception:
                n_sources_space = None

        if (
            n_sources_space is not None
            and n_leadfield_columns is not None
            and n_sources_space > 0
            and n_leadfield_columns % n_sources_space == 0
        ):
            n_orient = int(n_leadfield_columns // n_sources_space)

        if compact:
            # Minimal payload for the MkDocs dashboard: aggregated metrics only.
            output = {
                "ranks": per_dataset_ranks,
                "global_ranks": global_ranks,
                "metadata": {
                    "name": name,
                    "description": description,
                    "timestamp": datetime.now().isoformat(),
                    "n_samples": self.n_samples,
                    "random_seed": self.random_seed,
                    "solvers": self.solvers,
                    "m": m_electrodes,
                    "n": n_sources_space,
                    "m_electrodes": m_electrodes,
                    "n_sources": n_sources_space,
                    "n_leadfield_columns": n_leadfield_columns,
                    "n_orient": n_orient,
                },
                "datasets": datasets_payload,
                "results": [
                    {
                        "solver_name": r.solver_name,
                        "dataset_name": r.dataset_name,
                        "category": r.category,
                        "metrics": {
                            metric: {
                                "mean": float(stats.mean),
                                "std": float(stats.std),
                                "median": float(stats.median),
                                "worst_10_pct": (
                                    float(stats.worst_10_pct)
                                    if stats.worst_10_pct is not None
                                    else None
                                ),
                            }
                            for metric, stats in r.metrics.items()
                        },
                        "timing": r.timing.model_dump() if r.timing else None,
                        "samples": [],
                    }
                    for r in self._results
                ],
            }
            path.write_text(json.dumps(output, indent=2))
            logger.info("Compact results saved to %s", path)
            return

        # Full payload (includes per-sample metrics)
        summary = {}
        for r in self._results:
            key = f"{r.solver_name} | {r.dataset_name}"
            summary[key] = {m: round(s.mean, 4) for m, s in r.metrics.items()}

        best_solvers = self._compute_best_solvers()
        output = {
            "summary": summary,
            "best_solvers": best_solvers,
            "ranks": per_dataset_ranks,
            "global_ranks": global_ranks,
            "metadata": {
                "name": name,
                "description": description,
                "timestamp": datetime.now().isoformat(),
                "n_samples": self.n_samples,
                "random_seed": self.random_seed,
                "solvers": self.solvers,
                "m": m_electrodes,
                "n": n_sources_space,
                "m_electrodes": m_electrodes,
                "n_sources": n_sources_space,
                "n_leadfield_columns": n_leadfield_columns,
                "n_orient": n_orient,
            },
            "datasets": datasets_payload,
            "results": [r.model_dump() for r in self._results],
        }
        path.write_text(json.dumps(output, indent=2))
        logger.info("Results saved to %s", path)

    @classmethod
    def load(cls, path: str | Path) -> list[BenchmarkResult]:
        path = Path(path)
        data = json.loads(path.read_text())
        results = []
        for r in data["results"]:
            # Populate category if missing (for backward compatibility)
            if "category" not in r or r["category"] is None:
                r["category"] = get_solver_category(r["solver_name"])
            # Add correlation if missing (backward compatibility)
            for sample in r.get("samples", []):
                if "correlation" not in sample:
                    sample["correlation"] = float("nan")
            results.append(BenchmarkResult(**r))
        return results

    @classmethod
    def update_summary_statistics(cls, path: str | Path) -> None:
        """Update summary statistics (including best_solvers) for an existing results file.

        This is useful when you want to regenerate the summary from existing results
        without re-running the benchmark.

        Parameters
        ----------
        path : str or Path
            Path to the benchmark results JSON file.
        """
        path = Path(path)
        data = json.loads(path.read_text())
        results = []
        for r in data["results"]:
            # Populate category if missing (for backward compatibility)
            if "category" not in r or r["category"] is None:
                r["category"] = get_solver_category(r["solver_name"])
            # Add correlation if missing (backward compatibility)
            for sample in r.get("samples", []):
                if "correlation" not in sample:
                    sample["correlation"] = float("nan")
            results.append(BenchmarkResult(**r))

        # Create a temporary runner instance to use the _compute_best_solvers method
        # We need to set _results manually
        temp_runner = cls.__new__(cls)
        temp_runner._results = results

        # Recompute summary
        summary = {}
        for r in results:
            key = f"{r.solver_name} | {r.dataset_name}"
            summary[key] = {m: round(s.mean, 4) for m, s in r.metrics.items()}

        # Compute best solvers and ranks
        best_solvers = temp_runner._compute_best_solvers()
        per_dataset_ranks, global_ranks = temp_runner._compute_average_ranks()

        # Update the data structure
        data["summary"] = summary
        data["best_solvers"] = best_solvers
        data["ranks"] = per_dataset_ranks
        data["global_ranks"] = global_ranks

        # Write back
        path.write_text(json.dumps(data, indent=2))
        logger.info("Updated summary statistics in %s", path)



"""OmniChampagne: an adaptive sparse Bayesian solver that tries to win everywhere.

Motivation (from benchmark observations):
- Focal (single/multi dipole): Champagne-style sparse Bayesian learning excels.
- Extended patches: FLEX/SSM-style multi-order dictionaries are needed.

This solver evaluates two related SBL models per sample:
1) Dipole model (order 0 only)  -> behaves like Champagne.
2) Patch model (multi-order diffusion basis matching the simulator) -> behaves like Flex-like patch solvers.

It picks the model with the best penalized evidence (negative log-likelihood +
complexity penalty), then returns the corresponding inverse operator in the
original source space.
"""

from __future__ import annotations

import logging
from dataclasses import dataclass

import mne
import numpy as np
from scipy.sparse import csr_matrix, vstack
from scipy.sparse.csgraph import laplacian

from invert.util import build_source_adjacency

from ..base import BaseSolver, InverseOperator, SolverMeta
from .sbl_core import sbl_iterate

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class _SBLFit:
    active_set: np.ndarray  # indices into atoms
    gammas: np.ndarray  # gammas for active atoms (len == len(active_set))
    sigma_y_inv: np.ndarray  # (n_chans, n_chans)
    loss: float


class SolverOmniChampagne(BaseSolver):
    """Adaptive Champagne/Flex-like solver via model selection."""

    meta = SolverMeta(
        slug="omni-champagne",
        full_name="OmniChampagne",
        category="Bayesian",
        description=(
            "Adaptive sparse Bayesian solver that selects between a dipole-only "
            "Champagne-style model and a multi-order patch-dictionary model based "
            "on penalized evidence."
        ),
        references=[
            "Lukas Hecker 2025, unpublished",
        ],
    )

    def __init__(
        self,
        name: str = "OmniChampagne",
        # Patch basis settings (match simulator defaults)
        n_orders: int = 3,
        diffusion_parameter: float = 0.1,
        adjacency_type: str = "spatial",
        adjacency_distance: float = 3e-3,
        # SBL settings
        update_rule: str = "MacKay",
        max_iter: int = 2000,
        pruning_thresh: float = 1e-3,
        convergence_criterion: float = 1e-8,
        # Model selection (penalize extra active atoms)
        complexity_penalty: float = 0.2,
        **kwargs,
    ) -> None:
        self.name = name
        self.n_orders = int(n_orders)
        self.diffusion_parameter = float(diffusion_parameter)
        self.adjacency_type = str(adjacency_type)
        self.adjacency_distance = float(adjacency_distance)

        self.update_rule = str(update_rule)
        self.max_iter = int(max_iter)
        self.pruning_thresh = float(pruning_thresh)
        self.convergence_criterion = float(convergence_criterion)

        self.complexity_penalty = float(complexity_penalty)

        super().__init__(**kwargs)

    # ------------------------------------------------------------------
    # Public API
    # ------------------------------------------------------------------
    def make_inverse_operator(  # type: ignore[override]
        self,
        forward,
        mne_obj,
        *args,
        alpha: str | float = "auto",
        noise_cov: mne.Covariance | None = None,
        **kwargs,
    ):
        super().make_inverse_operator(forward, mne_obj, *args, alpha=alpha, **kwargs)
        wf = self.prepare_whitened_forward(noise_cov)
        data = self.unpack_data_obj(mne_obj)
        data = wf.sensor_transform @ data

        inv_op = self._fit_and_build_inverse_operator(data)
        self.inverse_operators = [
            InverseOperator(inv_op @ wf.sensor_transform, self.name)
        ]
        return self

    # ------------------------------------------------------------------
    # Model selection + inverse construction
    # ------------------------------------------------------------------
    def _fit_and_build_inverse_operator(self, Y: np.ndarray) -> np.ndarray:
        n_chans, n_dipoles = self.leadfield.shape

        # Noise estimate from data covariance
        C_y = self.data_covariance(Y, center=True, ddof=1)
        alpha_noise = float(np.trace(C_y) / n_chans)
        noise_cov = alpha_noise * np.eye(n_chans)
        Y_scaled = Y

        # Model A: dipole-only dictionary
        L_dip = self.leadfield
        fit_dip = self._fit_sbl(
            L_dip,
            Y_scaled,
            noise_cov=noise_cov,
            max_iter=self.max_iter,
            pruning_thresh=self.pruning_thresh,
            conv_crit=self.convergence_criterion,
            update_rule=self.update_rule,
        )
        W_dip = self._inverse_from_fit(L_dip, fit_dip, noise_cov=noise_cov)

        # Model B: patch dictionary (multi-order diffusion basis)
        sources_full = self._build_simulator_basis(n_dipoles)
        # For the patch model we intentionally exclude the order-0 (identity)
        # basis because we evaluate a separate dipole-only model above.
        sources = (
            sources_full[n_dipoles:, :]
            if sources_full.shape[0] > n_dipoles
            else sources_full
        )
        L_patch = self.leadfield @ sources.T  # (n_chans, n_candidates)
        fit_patch = self._fit_sbl(
            L_patch,
            Y_scaled,
            noise_cov=noise_cov,
            max_iter=self.max_iter,
            pruning_thresh=self.pruning_thresh,
            conv_crit=self.convergence_criterion,
            update_rule=self.update_rule,
        )
        W_patch_coeff = self._inverse_from_fit(L_patch, fit_patch, noise_cov=noise_cov)
        # Map coefficients back to dipole space (sources.T is (n_dipoles, n_candidates))
        W_patch = (
            sources.T[:, fit_patch.active_set] @ W_patch_coeff[fit_patch.active_set]
        )

        # Penalized evidence for selection
        score_dip = fit_dip.loss + self.complexity_penalty * float(
            len(fit_dip.active_set)
        )
        score_patch = fit_patch.loss + self.complexity_penalty * float(
            len(fit_patch.active_set)
        )

        if score_patch < score_dip:
            return W_patch
        return W_dip

    # ------------------------------------------------------------------
    # Basis construction (match SimulationGenerator logic)
    # ------------------------------------------------------------------
    def _build_simulator_basis(self, n_dipoles: int) -> np.ndarray:
        """Return stacked multi-order basis S with shape (n_candidates, n_dipoles)."""
        if self.n_orders <= 1:
            return np.eye(n_dipoles)

        I = np.eye(n_dipoles)
        if self.adjacency_type == "spatial":
            adjacency = build_source_adjacency(
                self.forward["src"],
                adjacency_type="spatial",
                adjacency_distance=self.adjacency_distance,
                verbose=0,
            )
        else:
            adjacency = mne.spatial_dist_adjacency(
                self.forward["src"], self.adjacency_distance, verbose=None
            )
        adjacency = csr_matrix(adjacency)
        G = csr_matrix(I - self.diffusion_parameter * laplacian(adjacency))

        sources = csr_matrix(I)
        for _ in range(1, self.n_orders):
            last_block = sources.toarray()[-n_dipoles:, -n_dipoles:]
            new_sources = csr_matrix(last_block) @ G
            col_max = new_sources.max(axis=0).toarray().ravel()
            col_max = np.maximum(col_max, 1e-12)
            new_sources = new_sources / col_max[np.newaxis]
            sources = vstack([sources, new_sources])

        return sources.toarray()

    # ------------------------------------------------------------------
    # Sparse Bayesian learning core
    # ------------------------------------------------------------------
    def _fit_sbl(
        self,
        L_orig: np.ndarray,
        Y_scaled: np.ndarray,
        *,
        noise_cov: np.ndarray,
        max_iter: int,
        pruning_thresh: float,
        conv_crit: float,
        update_rule: str,
    ) -> _SBLFit:
        result = sbl_iterate(
            L=L_orig,
            Y=Y_scaled,
            noise_cov=noise_cov,
            update_rule=update_rule,
            max_iter=max_iter,
            pruning_thresh=pruning_thresh,
            conv_crit=conv_crit,
        )
        return _SBLFit(
            active_set=result.active_set.astype(int, copy=False),
            gammas=result.gammas.astype(float, copy=False),
            sigma_y_inv=result.Sigma_y_inv,
            loss=result.loss,
        )

    @staticmethod
    def _inverse_from_fit(
        L: np.ndarray, fit: _SBLFit, *, noise_cov: np.ndarray
    ) -> np.ndarray:
        """Build full inverse operator W (n_atoms, n_chans) for a given dictionary."""
        n_chans, n_atoms = L.shape

        gam_full = np.zeros(n_atoms, dtype=float)
        gam_full[fit.active_set] = fit.gammas
        Gamma = np.diag(gam_full)

        Sigma_y = noise_cov + (L * gam_full) @ L.T
        Sigma_y = 0.5 * (Sigma_y + Sigma_y.T)
        Sigma_y_inv = SolverOmniChampagne._robust_inv(Sigma_y)
        return Gamma @ L.T @ Sigma_y_inv  # (n_atoms, n_chans)

    @staticmethod
    def _robust_inv(M: np.ndarray) -> np.ndarray:
        try:
            return np.linalg.inv(M)
        except np.linalg.LinAlgError:
            return np.linalg.pinv(M)



import logging
from copy import deepcopy

import numpy as np

logger = logging.getLogger(__name__)


def soft_threshold(x, alpha):
    """Applies the soft thresholding operator to x with threshold alpha.

    Parameters
    ----------
    x : ndarray, shape (n,)
        Input array.
    alpha : float
        Threshold.

    Returns
    -------
    y : ndarray, shape (n,)
        Output array.
    """
    y = np.sign(x) * np.maximum(np.abs(x) - alpha, 0)
    return y


def calc_eloreta_D2(leadfield, noise_cov, alpha, stop_crit=0.005, verbose=0):
    """Algorithm that optimizes weight matrix D as described in
    Assessing interactions in the brain with exactlow-resolution electromagnetic tomography; Pascual-Marqui et al. 2011 and
    https://www.sciencedirect.com/science/article/pii/S1053811920309150
    """
    n_chans, n_dipoles = leadfield.shape
    # initialize weight matrix D with identity and some empirical shift (weights are usually quite smaller than 1)
    D = np.identity(n_dipoles)

    if verbose > 0:
        logger.info("Optimizing eLORETA weight matrix W...")
    cnt = 0
    while True:
        old_D = deepcopy(D)
        if verbose > 0:
            logger.debug(f"\trep {cnt + 1}")
        D_inv = np.linalg.inv(D)
        inner_term = np.linalg.inv(
            leadfield @ D_inv @ leadfield.T + alpha**2 * noise_cov
        )

        for v in range(n_dipoles):
            D[v, v] = np.sqrt(leadfield[:, v].T @ inner_term @ leadfield[:, v])

        averagePercentChange = np.abs(
            1 - np.mean(np.divide(np.diagonal(D), np.diagonal(old_D)))
        )

        if verbose > 0:
            logger.debug(f"averagePercentChange={100 * averagePercentChange:.2f} %")

        if averagePercentChange < stop_crit:
            if verbose > 0:
                logger.info("...converged...")
            break
        cnt += 1
    if verbose > 0:
        logger.info("...done!")
    return D



import copy
import logging
from pathlib import Path

import mne
import numpy as np
from scipy.sparse import coo_matrix

logger = logging.getLogger(__name__)


class Parcellator:
    def __init__(
        self,
        forward: mne.Forward,
        adjacency: coo_matrix,
        subjects_dir: str | None = None,
    ):
        """
        Initialize the Parcellator.

        Parameters
        ----------
        forward : mne.Forward
            The forward solution.
        adjacency : coo_matrix
            The adjacency matrix.
        subjects_dir : str, optional
            Path to the subjects directory. If None, uses MNE's default.
        """
        self.forward = forward
        self.adjacency = adjacency
        self.leadfield = forward["sol"]["data"]
        self.leadfield_whitened = self.whiten(self.leadfield)
        self.hemispheres = ["lh", "rh"]
        self.B = None  # Will be set after parcellation
        self.forward_parcellated = None
        self.parcel_adjacency = None  # Will be set after parcellation

        if subjects_dir is None:
            self.subjects_dir = Path(mne.datasets.sample.data_path()) / "subjects"
        else:
            self.subjects_dir = subjects_dir  # type: ignore[assignment]

        self.src = self.forward["src"]
        self.vertices = [self.src[0]["vertno"], self.src[1]["vertno"]]

    def parcellate(
        self, parcellation: str = "aparc.a2009s", subject: str | None = None
    ):
        """
        Parcellate the leadfield using the given parcellation scheme.

        Parameters
        ----------
        parcellation : str
            The parcellation to use (default: "aparc.a2009s").
        subject : str, optional
            Subject name. If None, extracts from forward solution.

        Returns
        -------
        fwd_parcellated : mne.Forward
            Parcellated forward solution.

        """
        # Extract subject from forward solution if not provided
        if subject is None:
            subject = self.forward["src"][0].get("subject_his_id", "fsaverage")

        # Read labels for both hemispheres and flatten the list
        labels = []
        for hemisphere in self.hemispheres:
            #     hemi_labels = mne.read_labels_from_annot(
            #         subject,
            #         parc=parcellation,
            #         hemi=hemisphere,
            #         subjects_dir=self.subjects_dir
            #     )
            hemi_labels = mne.read_labels_from_annot(
                subject, parcellation, hemisphere, subjects_dir=self.subjects_dir
            )
            labels.extend(hemi_labels)

        # Build parcellation basis and get representative vertices for each parcel
        self.B, parcel_vertices_lh, parcel_vertices_rh = (
            self._parcellation_basis_from_labels(labels)
        )

        # Compute parcellated leadfield
        L_B = self.leadfield @ self.B

        # Compute parcel-to-parcel adjacency
        self.parcel_adjacency = self._compute_parcel_adjacency()

        # Create a copy of the forward solution and modify its leadfield
        fwd_parcellated = copy.deepcopy(self.forward)
        fwd_parcellated["sol"]["data"] = L_B
        if self.B is not None:
            fwd_parcellated["nsource"] = self.B.shape[1]
        fwd_parcellated["source_nn"] = np.eye(3)[
            :, [2]
        ]  # Simplified, as parcels don't have orientations

        # Update the source space vertices to match the parcellated space
        fwd_parcellated["src"][0]["vertno"] = parcel_vertices_lh
        fwd_parcellated["src"][0]["nuse"] = len(parcel_vertices_lh)
        if len(fwd_parcellated["src"]) > 1:
            fwd_parcellated["src"][1]["vertno"] = parcel_vertices_rh
            fwd_parcellated["src"][1]["nuse"] = len(parcel_vertices_rh)

        # Store for later use
        self.forward_parcellated = fwd_parcellated
        self.parcel_vertices = [parcel_vertices_lh, parcel_vertices_rh]

        return fwd_parcellated

    def compress(self, stc: mne.SourceEstimate) -> mne.SourceEstimate:
        """
        Compress the source estimate using the parcel basis.

        Parameters
        ----------
        stc : mne.SourceEstimate
            The source estimate (n_dipoles x n_timepoints).

        Returns
        -------
        stc_parcellated : mne.SourceEstimate
            The compressed source estimate (n_parcels x n_timepoints).
        """
        if self.B is None:
            raise ValueError("Must call parcellate() before compress()")
        if not hasattr(self, "parcel_vertices"):
            raise ValueError(
                "parcel_vertices not found. Make sure parcellate() completed successfully."
            )

        X = stc.data
        X_parcellated = self.compress_data(X)
        stc_parcellated = mne.SourceEstimate(
            X_parcellated,
            vertices=self.parcel_vertices,
            tmin=stc.tmin,
            tstep=stc.tstep,
            subject=stc.subject,
        )
        return stc_parcellated

    def compress_data(self, data: np.ndarray) -> np.ndarray:
        """
        Compress the data using the parcel basis.
        """
        assert self.B is not None
        return self.B.T @ data

    def decompress(self, stc_parcellated: mne.SourceEstimate) -> mne.SourceEstimate:
        """
        Decompress the source estimate from the parcel basis.

        Parameters
        ----------
        stc_parcellated : mne.SourceEstimate
            The compressed parcel data (n_parcels x n_timepoints).

        Returns
        -------
        stc : mne.SourceEstimate
            The decompressed source estimate (n_dipoles x n_timepoints).
        """
        if self.B is None:
            raise ValueError("Must call parcellate() before decompress()")

        X_parcellated = stc_parcellated.data
        X = self.decompress_data(
            X_parcellated
        )  # Project from parcel space back to dipole space
        stc = mne.SourceEstimate(
            X,
            vertices=self.vertices,
            tmin=stc_parcellated.tmin,
            tstep=stc_parcellated.tstep,
            subject=stc_parcellated.subject,
        )
        return stc

    def decompress_data(self, data: np.ndarray) -> np.ndarray:
        """
        Decompress the data using the parcel basis.
        """
        return self.B @ data

    def get_parcel_adjacency(self) -> coo_matrix:
        """
        Get the parcel-to-parcel adjacency matrix.

        Returns
        -------
        parcel_adjacency : coo_matrix
            Sparse adjacency matrix for parcels (n_parcels x n_parcels).

        Raises
        ------
        ValueError
            If parcellate() has not been called yet.
        """
        if self.parcel_adjacency is None:
            raise ValueError("Must call parcellate() before getting parcel adjacency")
        return self.parcel_adjacency

    def _parcellation_basis_from_labels(self, labels: list) -> tuple:
        """
        Build a parcel basis matrix B (n_dipoles x n_labels)
        mapping dipoles in the forward model to parcellation regions.

        Each column corresponds to one label (ROI), normalized to unit L2 norm.
        Also returns representative vertices for each parcel.

        Parameters
        ----------
        labels : list of mne.Label
            List of labels defining the parcellation.

        Returns
        -------
        B : np.ndarray
            Parcellation basis matrix (n_dipoles x n_labels).
        parcel_vertices_lh : np.ndarray
            Representative vertices for left hemisphere parcels (sorted).
        parcel_vertices_rh : np.ndarray
            Representative vertices for right hemisphere parcels (sorted).
        """
        src = self.forward["src"]
        lh = src[0]["vertno"]
        rh = src[1]["vertno"] if len(src) > 1 else np.array([], int)
        n_dip = len(lh) + len(rh)

        # Create a lookup: dipole index in full source space -> position in B
        offset_rh = len(lh)

        # First pass: collect parcel info for each hemisphere
        lh_parcels = []  # List of (representative_vertex, label, label_idx)
        rh_parcels = []

        for k, label in enumerate(labels):
            if label.hemi == "lh":
                verts = np.intersect1d(label.vertices, lh)
                if len(verts) > 0:
                    lh_parcels.append((verts[0], label, k, verts))
            elif label.hemi == "rh":
                verts = np.intersect1d(label.vertices, rh)
                if len(verts) > 0:
                    rh_parcels.append((verts[0], label, k, verts))

        # Sort by representative vertex to ensure vertices are in increasing order
        lh_parcels.sort(key=lambda x: x[0])
        rh_parcels.sort(key=lambda x: x[0])

        # Build B matrix in sorted order
        n_parcels = len(lh_parcels) + len(rh_parcels)
        B = np.zeros((n_dip, n_parcels))

        parcel_verts_lh = []
        parcel_verts_rh = []

        # Fill in left hemisphere parcels
        parcel_idx = 0
        for rep_vert, _label, _orig_idx, verts in lh_parcels:
            parcel_verts_lh.append(rep_vert)
            idx = np.searchsorted(lh, verts)
            B[idx, parcel_idx] = 1.0 / np.sqrt(len(idx))
            parcel_idx += 1

        # Fill in right hemisphere parcels
        for rep_vert, _label, _orig_idx, verts in rh_parcels:
            parcel_verts_rh.append(rep_vert)
            idx = np.searchsorted(rh, verts) + offset_rh
            B[idx, parcel_idx] = 1.0 / np.sqrt(len(idx))
            parcel_idx += 1

        return (
            B,
            np.array(parcel_verts_lh, dtype=int),
            np.array(parcel_verts_rh, dtype=int),
        )

    def _compute_parcel_adjacency(self) -> coo_matrix:
        """
        Compute parcel-to-parcel adjacency based on the original source space adjacency.

        Two parcels are considered adjacent if any of their constituent vertices
        are adjacent in the original source space.

        Returns
        -------
        parcel_adj : coo_matrix
            Sparse adjacency matrix for parcels (n_parcels x n_parcels).
        """
        if self.B is None:
            raise ValueError("Must build parcellation basis first")

        n_parcels = self.B.shape[1]
        n_vertices = self.B.shape[0]

        # Build vertex-to-parcel mapping: which parcel does each vertex belong to?
        vertex_to_parcel = np.full(n_vertices, -1, dtype=int)
        for vertex_idx in range(n_vertices):
            parcel_indices = np.nonzero(self.B[vertex_idx, :])[0]
            if len(parcel_indices) > 0:
                vertex_to_parcel[vertex_idx] = parcel_indices[0]

        # Make adjacency symmetric to capture all edges (handles both directions)
        adj_csr = self.adjacency.tocsr()
        adj_symmetric = adj_csr + adj_csr.T
        adj_coo = adj_symmetric.tocoo()

        # Collect parcel edges
        parcel_edges = set()
        for i, j in zip(adj_coo.row, adj_coo.col, strict=False):
            # Skip if either vertex is not in any parcel (shouldn't happen, but be safe)
            if vertex_to_parcel[i] < 0 or vertex_to_parcel[j] < 0:
                continue

            parcel_i = vertex_to_parcel[i]
            parcel_j = vertex_to_parcel[j]

            # Only add if different parcels
            if parcel_i != parcel_j:
                parcel_edges.add((parcel_i, parcel_j))
                parcel_edges.add((parcel_j, parcel_i))

        # Convert to COO matrix
        if len(parcel_edges) > 0:
            edges = np.array(list(parcel_edges))
            row = edges[:, 0]
            col = edges[:, 1]
            data = np.ones(len(row))
            parcel_adj = coo_matrix((data, (row, col)), shape=(n_parcels, n_parcels))
        else:
            parcel_adj = coo_matrix((n_parcels, n_parcels))

        # Debug: Check if any parcels have no neighbors
        parcel_degrees = np.array(parcel_adj.sum(axis=1)).flatten()
        isolated_parcels = np.where(parcel_degrees == 0)[0]
        if len(isolated_parcels) > 0:
            logger.warning(
                f"{len(isolated_parcels)} parcel(s) have no neighbors: {isolated_parcels}"
            )
            # This could indicate vertices with no edges in the adjacency matrix
            # or parcels at the mesh boundary

        return parcel_adj

    def diagnose_parcel_connectivity(self) -> dict:
        """
        Diagnose connectivity issues in parcel adjacency.

        Returns
        -------
        diagnostics : dict
            Contains:
            - 'n_parcels': number of parcels
            - 'n_with_neighbors': number of parcels with at least one neighbor
            - 'n_isolated': number of parcels with no neighbors
            - 'isolated_parcel_ids': list of isolated parcel indices
            - 'parcel_degrees': array of neighbor counts per parcel
            - 'isolated_vertices_per_parcel': dict mapping parcel id to list of isolated vertex indices
        """
        if self.B is None:
            raise ValueError(
                "Must call parcellate() before diagnose_parcel_connectivity()"
            )

        n_parcels = self.B.shape[1]
        n_vertices = self.B.shape[0]

        # Get adjacency info
        adj_csr = self.adjacency.tocsr()
        adj_symmetric = adj_csr + adj_csr.T

        # Vertex-to-parcel mapping
        vertex_to_parcel = np.full(n_vertices, -1, dtype=int)
        for vertex_idx in range(n_vertices):
            parcel_indices = np.nonzero(self.B[vertex_idx, :])[0]
            if len(parcel_indices) > 0:
                vertex_to_parcel[vertex_idx] = parcel_indices[0]

        # Count neighbors for each vertex
        vertex_degrees = np.array(adj_symmetric.sum(axis=1)).flatten()

        # Find isolated vertices per parcel
        isolated_verts_per_parcel = {}
        for parcel_id in range(n_parcels):
            # Find all vertices in this parcel
            vertex_indices = np.where(vertex_to_parcel == parcel_id)[0]
            # Find which of these have no neighbors (degree == 0)
            isolated = vertex_indices[vertex_degrees[vertex_indices] == 0]
            if len(isolated) > 0:
                isolated_verts_per_parcel[parcel_id] = isolated.tolist()

        # Get parcel degrees from adjacency
        parcel_adj = (
            self.get_parcel_adjacency()
            if self.parcel_adjacency is not None
            else self._compute_parcel_adjacency()
        )
        parcel_degrees = np.array(parcel_adj.sum(axis=1)).flatten()

        isolated_parcels = np.where(parcel_degrees == 0)[0]

        return {
            "n_parcels": n_parcels,
            "n_with_neighbors": len(parcel_degrees) - len(isolated_parcels),
            "n_isolated": len(isolated_parcels),
            "isolated_parcel_ids": isolated_parcels.tolist(),
            "parcel_degrees": parcel_degrees,
            "isolated_vertices_per_parcel": isolated_verts_per_parcel,
        }

    @staticmethod
    def whiten(leadfield: np.ndarray) -> np.ndarray:
        """
        Whiten the leadfield by normalizing each column.

        Parameters
        ----------
        leadfield : np.ndarray
            The leadfield matrix (n_sensors x n_dipoles).

        Returns
        -------
        leadfield_whitened : np.ndarray
            The whitened leadfield matrix.
        """
        return leadfield / np.linalg.norm(leadfield, axis=0, keepdims=True)

    @staticmethod
    def trim_to_roi(
        fwd: mne.Forward,
        parcellation: str = "HCPMMP1_combined",
        roi: str = "front",
        subject: str = "fsaverage",
        subjects_dir: str | None = None,
        verbose: bool = True,
    ) -> mne.Forward:
        """
        Trim a forward solution to only include vertices in a specific ROI.

        This is useful for computing beamformers or other inverse solutions
        only in regions of interest.

        Parameters
        ----------
        fwd : mne.Forward
            The forward solution to trim.
        parcellation : str, optional
            The parcellation to use (default: "HCPMMP1_combined").
        roi : str, optional
            ROI specification. Labels containing this string (case-insensitive)
            will be included. Examples: 'front' for frontal lobe, 'temporal'
            for temporal lobe, 'precentral' for motor cortex (default: "front").
        subject : str, optional
            Subject name (default: "fsaverage").
        subjects_dir : str, optional
            Path to the subjects directory. If None, uses MNE's default.
        verbose : bool, optional
            Whether to print information about the trimming (default: True).

        Returns
        -------
        fwd_roi : mne.Forward
            Trimmed forward solution containing only vertices in the ROI.

        Raises
        ------
        ValueError
            If no labels matching the ROI are found.
            If no vertices in the forward solution belong to the ROI.
        """
        # Load parcellation labels
        labels_lh = mne.read_labels_from_annot(
            subject,
            parc=parcellation,
            hemi="lh",
            subjects_dir=subjects_dir,
            verbose=False,
        )
        labels_rh = mne.read_labels_from_annot(
            subject,
            parc=parcellation,
            hemi="rh",
            subjects_dir=subjects_dir,
            verbose=False,
        )
        all_labels = labels_lh + labels_rh

        # Find labels matching the ROI specification
        roi_labels = [
            label for label in all_labels if roi.lower() in label.name.lower()
        ]

        return mne.forward.restrict_forward_to_label(fwd, roi_labels)



import logging

import mne
import numpy as np

from ..base import SolverMeta
from .base_beamformer import BaseBeamformer

logger = logging.getLogger(__name__)


class SolverSSPESMV(BaseBeamformer):
    """Hybrid Simultaneous Subspace Pursuit + Eigenspace Minimum Variance solver.

    Combines SSP's greedy sparse support identification with ESMV's
    eigenspace-projected beamforming for amplitude recovery. This targets
    the highly underdetermined scenario (e.g., 4-channel Muse headband)
    where localization accuracy (SSP's strength) and amplitude fidelity
    (ESMV's strength) are both critical.

    Algorithm
    ---------
    1. Normalize leadfield columns for unbiased atom selection.
    2. Run SSP iterations to identify the sparse support set T.
    3. Build a reduced leadfield from the support set.
    4. Apply ESMV-style eigenspace beamforming on the reduced problem
       to get accurate amplitude estimates.

    References
    ----------
    SSP: Dai & Milenkovic (2009). Subspace pursuit for compressive sensing.
    ESMV: Jonmohamadi et al. (2014). Comparison of beamformers for EEG.
    """

    meta = SolverMeta(
        slug="ssp_esmv",
        full_name="Subspace Pursuit + ESMV",
        category="Beamformers",
        description=(
            "Hybrid solver that uses Subspace Pursuit to identify a sparse support "
            "set, then applies ESMV-style beamforming on the reduced problem."
        ),
        references=[
            "Lukas Hecker (2025). Unpublished.",
            "Dai, W., & Milenkovic, O. (2009). Subspace pursuit for compressive sensing "
            "signal reconstruction. IEEE Transactions on Information Theory, 55(5), "
            "2230-2249.",
            "Jonmohamadi, Y., Poudel, G., Innes, C., Weiss, D., Krueger, R., & Jones, R. "
            "(2014). Comparison of beamformers for EEG source signal reconstruction. "
            "Biomedical Signal Processing and Control, 14, 175-188.",
        ],
    )

    def __init__(self, name="SSP-ESMV", reduce_rank=True, rank="auto", **kwargs):
        self.name = name
        return super().__init__(reduce_rank=reduce_rank, rank=rank, **kwargs)

    def make_inverse_operator(
        self,
        forward,
        mne_obj=None,
        *args,
        alpha="auto",
        noise_cov: mne.Covariance | None = None,
        **kwargs,
    ):
        super().make_inverse_operator(forward, mne_obj, *args, alpha=alpha, **kwargs)
        self.prepare_whitened_forward(noise_cov)
        self.leadfield_original = self.leadfield.copy()
        self.leadfield_normed = self.robust_normalize_leadfield(self.leadfield)
        self.inverse_operators = []
        return self

    def apply_inverse_operator(self, mne_obj) -> mne.SourceEstimate:
        data = self.unpack_data_obj(mne_obj)
        self.validate_operator_data_compatibility(data)
        data = self._sensor_transform @ data
        source_mat = self._solve(data)
        stc = self.source_to_object(source_mat)
        return stc

    def _solve(self, y):
        n_chans, n_time = y.shape
        leadfield = self.leadfield_original
        leadfield_normed = self.leadfield_normed
        _, n_dipoles = leadfield.shape

        K = max(1, n_chans // 2)

        # --- Phase 1: SSP support identification ---
        b = np.linalg.norm(leadfield_normed.T @ y, axis=1)
        T = self._top_k_indices(b, K)
        R = self._residual(y, leadfield[:, T])

        T_prev = T
        for _i in range(n_chans):
            b = np.linalg.norm(leadfield_normed.T @ R, axis=1)
            new_atoms = self._top_k_indices(b, K)
            T_tilde = np.unique(np.concatenate([T_prev, new_atoms]))

            if len(T_tilde) > 0:
                xp = self.robust_inverse_solution(leadfield[:, T_tilde], y)
                xp_norm = np.linalg.norm(xp, axis=1)
                T_candidate = T_tilde[self._top_k_indices(xp_norm, K)]
            else:
                T_candidate = T_tilde

            if len(T_candidate) > 0:
                xp_final = self.robust_inverse_solution(leadfield[:, T_candidate], y)
                R_new = y - leadfield[:, T_candidate] @ xp_final
            else:
                R_new = y

            if np.linalg.norm(R_new) >= np.linalg.norm(R):
                break

            R = R_new
            T_prev = T_candidate
            T = T_candidate

        # --- Phase 2: ESMV amplitude recovery on the identified support ---
        x_hat = np.zeros((n_dipoles, n_time))
        if len(T) == 0:
            return x_hat

        L_T = leadfield[:, T]
        L_T_norm = L_T / np.linalg.norm(L_T, axis=0)

        # Eigenspace beamformer on the reduced problem
        C = y @ y.T
        C_sub = self.select_signal_subspace(C)

        eigs = np.linalg.svd(C, compute_uv=False)
        alpha_val = eigs.max() * 1e-3  # moderate regularization

        I = np.eye(n_chans)
        C_inv = self.robust_inverse(C + alpha_val * I)
        C_inv_L = C_inv @ L_T_norm
        diag = np.einsum("ij,ji->i", L_T_norm.T, C_inv_L)
        diag = np.maximum(diag, 1e-12)
        W = C_inv_L / diag
        W_esmv = C_sub @ W

        x_hat[T] = W_esmv.T @ y

        return x_hat

    @staticmethod
    def _top_k_indices(values, k):
        k = min(k, len(values))
        return np.argpartition(values, -k)[-k:]

    def _residual(self, y, L_sub):
        if L_sub.shape[1] == 0:
            return y.copy()
        return self.robust_residual(y, L_sub)



import logging

import mne
import numpy as np

from ..base import InverseOperator, SolverMeta
from .base_beamformer import BaseBeamformer
from .utils import build_covariance_candidates

logger = logging.getLogger(__name__)


class SolverESMV(BaseBeamformer):
    """Class for the Eigenspace-based Minimum Variance (ESMV) Beamformer
        inverse solution [1].

    References
    ----------
    [1] Jonmohamadi, Y., Poudel, G., Innes, C., Weiss, D., Krueger, R., & Jones,
    R. (2014). Comparison of beamformers for EEG source signal reconstruction.
    Biomedical Signal Processing and Control, 14, 175-188.

    """

    meta = SolverMeta(
        slug="esmv",
        full_name="Eigenspace-based Minimum Variance",
        category="Beamformers",
        description=(
            "Eigenspace-projected minimum-variance beamformer (ESMV), using a "
            "signal-subspace projection of the classic MV filter."
        ),
        references=[
            "Jonmohamadi, Y., Poudel, G., Innes, C., Weiss, D., Krueger, R., & Jones, R. "
            "(2014). Comparison of beamformers for EEG source signal reconstruction. "
            "Biomedical Signal Processing and Control, 14, 175-188.",
        ],
    )
    R_VALUE_EXPONENTS = (-16.0, 1.0)

    def __init__(
        self,
        name="ESMV Beamformer",
        reduce_rank=True,
        rank="auto",
        use_robust_covariance: bool = False,
        use_noise_whitening: bool = True,
        subspace_weighting: str = "eig",
        enforce_unit_gain_after_subspace: bool = False,
        rank_tol: float = 1e-12,
        eps: float = 1e-15,
        **kwargs,
    ):
        self.name = name
        self.use_robust_covariance = bool(use_robust_covariance)
        self.use_noise_whitening = bool(use_noise_whitening)
        self.subspace_weighting = str(subspace_weighting).lower()
        self.enforce_unit_gain_after_subspace = bool(enforce_unit_gain_after_subspace)
        self.rank_tol = float(rank_tol)
        self.eps = float(eps)
        return super().__init__(reduce_rank=reduce_rank, rank=rank, **kwargs)

    def make_inverse_operator(
        self,
        forward,
        mne_obj=None,
        *args,
        alpha="auto",
        noise_cov: mne.Covariance | None = None,
        cov_reg: str = "oas",
        cov_reg_beta: float = 0.05,
        cov_reg_cond_target: float = 1e4,
        **kwargs,
    ):
        """Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        mne_obj : [mne.Evoked, mne.Epochs, mne.io.Raw]
            The MNE data object.
        alpha : float
            The regularization parameter.

        Return
        ------
        self : object returns itself for convenience

        """
        super().make_inverse_operator(forward, mne_obj, *args, alpha=alpha, **kwargs)
        if noise_cov is not None:
            self.coerce_noise_cov(noise_cov)
        # For beamformers, covariance estimation should use many samples.
        # If Epochs are provided, concatenate epochs rather than averaging.
        data, kept = self.unpack_covariance_samples(
            mne_obj,
            pick_types=["meg", "eeg", "fnirs"],
            require_forward_channel_match=True,
            context="ESMV",
        )
        self._operator_ch_names = tuple(kept)

        leadfield = self.leadfield
        _n_chans_raw, _n_dipoles = leadfield.shape
        y_raw = np.asarray(data, dtype=float)

        lead_norms = np.linalg.norm(leadfield, axis=0)
        leadfield /= np.maximum(lead_norms, self.eps)

        if self.use_robust_covariance:
            if self.use_noise_whitening:
                if noise_cov is None:
                    noise_cov = self.make_identity_noise_cov(
                        list(self.forward.ch_names)
                    )
                wf = self.prepare_whitened_forward(
                    noise_cov,
                    rank_tol=self.rank_tol,
                    eps=self.eps,
                )
                if wf.whitener_mode not in ("projected", "none"):
                    logger.warning("ESMV whitener fallback used: %s", wf.whitener_mode)
            else:
                wf = self.prepare_whitened_forward(None)
            sensor_transform = wf.sensor_transform
            leadfield_eff = wf.G_white
            y_eff = sensor_transform @ y_raw
        else:
            sensor_transform = None
            leadfield_eff = leadfield
            y_eff = y_raw
        n_chans_eff = int(leadfield_eff.shape[0])
        I = np.identity(n_chans_eff)

        # Recompute regularization based on the max eigenvalue of the Covariance
        # Matrix (opposed to that of the leadfield)
        y_eff -= y_eff.mean(axis=1, keepdims=True)
        C = self.data_covariance(y_eff, center=False, ddof=1)
        eigvals, eigvecs = np.linalg.eigh(C)
        order = np.argsort(eigvals)[::-1]
        eigvecs = eigvecs[:, order]
        if isinstance(self.rank, str):
            n_comp = self.estimate_n_sources(C, method=self.rank)
        elif isinstance(self.rank, int):
            n_comp = self.rank
        else:
            raise ValueError(
                f"Unknown rank: {self.rank}. Must be 'auto', 'L', 'drop', 'mean', or a positive integer."
            )
        n_comp = int(np.clip(n_comp, 1, n_chans_eff))
        signal_subspace = eigvecs[:, :n_comp]
        signal_eigs = np.maximum(eigvals[order][:n_comp], 0.0)
        if self.subspace_weighting == "projector":
            subspace_weights = np.ones(n_comp, dtype=float)
        elif self.subspace_weighting == "sqrt_eig":
            max_ev = max(float(np.max(signal_eigs)), self.eps)
            subspace_weights = np.sqrt(signal_eigs / max_ev)
        elif self.subspace_weighting == "eig":
            max_ev = max(float(np.max(signal_eigs)), self.eps)
            subspace_weights = signal_eigs / max_ev
        else:
            raise ValueError(
                "subspace_weighting must be one of {'projector', 'sqrt_eig', 'eig'}"
            )
        subspace_operator = signal_subspace @ (
            subspace_weights[:, np.newaxis] * signal_subspace.T
        )
        cov_mats, self.alphas, cov_meta = build_covariance_candidates(
            C=C,
            I=I,
            alpha=self.alpha,
            get_alphas_fn=self.get_alphas,
            n_samples=int(y_eff.shape[1]),
            cov_reg=cov_reg,
            cov_reg_beta=float(cov_reg_beta),
            cov_reg_cond_target=float(cov_reg_cond_target),
        )
        if "oas_shrinkage" in cov_meta:
            self._cov_reg_oas_shrinkage = float(cov_meta["oas_shrinkage"])

        inverse_operators = []
        for cov_mat in cov_mats:
            C_inv = self.robust_inverse(cov_mat)
            C_inv_leadfield = C_inv @ leadfield_eff
            diag_elements = np.einsum("ij,ji->i", leadfield_eff.T, C_inv_leadfield)
            W_mv = np.zeros_like(C_inv_leadfield)
            valid = np.abs(diag_elements) > self.eps
            if np.any(valid):
                W_mv[:, valid] = C_inv_leadfield[:, valid] / diag_elements[valid]
            W_esmv_eff = subspace_operator @ W_mv

            if self.enforce_unit_gain_after_subspace:
                gains = np.einsum("ij,ji->i", leadfield_eff.T, W_esmv_eff)
                valid_gain = np.abs(gains) > self.eps
                if np.any(valid_gain):
                    W_esmv_eff[:, valid_gain] = (
                        W_esmv_eff[:, valid_gain] / gains[valid_gain]
                    )
                if np.any(~valid_gain):
                    W_esmv_eff[:, ~valid_gain] = W_mv[:, ~valid_gain]

            inverse_operator = W_esmv_eff.T
            # If we computed the filter in whitened/projected sensor space,
            # map it back so the operator can be applied to raw sensor data.
            if sensor_transform is not None:
                inverse_operator = inverse_operator @ sensor_transform
            inverse_operators.append(inverse_operator)

        self.inverse_operators = [
            InverseOperator(inverse_operator, self.name)
            for inverse_operator in inverse_operators
        ]
        return self



"""
Beamformer solvers for M/EEG source reconstruction.
"""

from .adapt_flex_esmv import SolverAdaptFlexESMV
from .base_beamformer import BaseBeamformer
from .deblur_flex_esmv import SolverDeblurFlexESMV
from .dics import SolverDICS
from .ebb import SolverEBB
from .esmv import SolverESMV
from .esmv_mvpure import SolverESMVMVPURE
from .flex_esmv import SolverFlexESMV
from .flex_esmv_mvpure import SolverFlexESMVMVPURE
from .hocmcmv import SolverHOCMCMV
from .hocmv import SolverHOCMV
from .iresmv import SolverIRESMV
from .lcmv import SolverLCMV
from .lcmv_mvpure import SolverLCMVMVPURE
from .mcmv import SolverMCMV
from .mvab import SolverMVAB
from .recipsiicos_plain import SolverReciPSIICOSPlain
from .recipsiicos_whitened import SolverReciPSIICOSWhitened
from .safe_flex_esmv import SolverSafeFlexESMV
from .sam import SolverSAM
from .sharp_flex_esmv import SolverSharpFlexESMV
from .smv import SolverSMV
from .ssp_esmv import SolverSSPESMV
from .ssp_iresmv import SolverSSPIRESMV
from .unit_noise_gain import SolverUnitNoiseGain
from .wnmv import SolverWNMV

__all__ = [
    "BaseBeamformer",
    "SolverMVAB",
    "SolverLCMV",
    "SolverLCMVMVPURE",
    "SolverDICS",
    "SolverSMV",
    "SolverWNMV",
    "SolverHOCMV",
    "SolverESMV",
    "SolverESMVMVPURE",
    "SolverAdaptFlexESMV",
    "SolverFlexESMV",
    "SolverFlexESMVMVPURE",
    "SolverDeblurFlexESMV",
    "SolverSafeFlexESMV",
    "SolverSharpFlexESMV",
    "SolverMCMV",
    "SolverUnitNoiseGain",
    "SolverHOCMCMV",
    "SolverSAM",
    "SolverEBB",
    "SolverReciPSIICOSPlain",
    "SolverReciPSIICOSWhitened",
    "SolverSSPESMV",
    "SolverIRESMV",
    "SolverSSPIRESMV",
]



from __future__ import annotations

import logging
from copy import deepcopy

import mne
import numpy as np

_TORCH_IMPORT_ERROR: ModuleNotFoundError | None = None
try:
    import torch  # type: ignore[import-not-found]
    import torch.nn as nn  # type: ignore[import-not-found]
except ModuleNotFoundError as exc:  # pragma: no cover
    torch = None  # type: ignore[assignment]

    class _NNModule:  # pragma: no cover
        pass

    class _NN:  # pragma: no cover
        Module = _NNModule

    nn = _NN()  # type: ignore[assignment]
    _TORCH_IMPORT_ERROR = exc

from ...simulate import SimulationGenerator
from ..base import BaseSolver, SolverMeta
from .torch_utils import (
    activation_from_name,
    count_trainable_parameters,
    get_torch_device,
    loss_from_name,
)
from .utils import rescale_sources

logger = logging.getLogger(__name__)


class _LSTMNet(nn.Module):
    def __init__(
        self,
        n_channels: int,
        n_dipoles: int,
        *,
        n_dense_units: int,
        n_lstm_units: int,
        activation_function: str,
    ) -> None:
        super().__init__()
        self.fc = nn.Linear(int(n_channels), int(n_dense_units))
        self.activation = activation_from_name(activation_function)
        self.direct_out = nn.Linear(int(n_dense_units), int(n_dipoles))
        self.lstm = nn.LSTM(
            input_size=int(n_dense_units),
            hidden_size=int(n_lstm_units),
            batch_first=True,
            bidirectional=True,
        )
        self.mask_out = nn.Linear(int(n_lstm_units) * 2, int(n_dipoles))
        self.mask_activation = nn.Sigmoid()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # x: (batch, time, channels)
        x = self.activation(self.fc(x))
        direct = self.direct_out(x)
        lstm_out, _state = self.lstm(x)
        mask = self.mask_activation(self.mask_out(lstm_out))
        return direct * mask


class SolverLSTM(BaseSolver):
    """Class for the Long-Short Term Memory Neural Network (LSTM) for
    EEG inverse solutions.

    """

    meta = SolverMeta(
        acronym="LSTM",
        full_name="Long Short-Term Memory Network",
        category="Neural Networks",
        description=(
            "Supervised recurrent (LSTM) network trained on simulated data to map "
            "sensor time series to source activity."
        ),
        references=["Lukas Hecker 2025, unpublished"],
    )

    def __init__(self, name="LSTM", **kwargs):
        self.name = name
        self.model = None
        self.optimizer = None
        self.device = None
        return super().__init__(**kwargs)

    def make_inverse_operator(
        self,
        forward,
        simulation_config=None,
        *args,
        n_dense_units=300,
        n_lstm_units=75,
        activation_function="tanh",
        epochs=300,
        learning_rate=1e-3,
        loss="cosine_similarity",
        size_validation_set=256,
        patience=100,
        noise_cov: mne.Covariance | None = None,
        alpha="auto",
        **kwargs,
    ):
        """Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        simulation_config : SimulationConfig
            A SimulationConfig object for data generation containing all
            simulation parameters (batch_size, n_sources, n_orders, snr_range, etc.).
        n_dense_units : int
            The number of neurons in the fully-connected hidden layers.
            Default 300.
        n_lstm_units : int
            The number of neurons in the LSTM hidden layers.
            Default 75.
        activation_function : str
            The activation function of the hidden layers.
            Default "tanh".
        epochs : int
            The number of epochs to train.
            Default 300.
        learning_rate : float
            The learning rate of the optimizer that trains the neural network.
            Default 1e-3.
        loss : str
            The loss function of the neural network.
            Default "cosine_similarity".
        size_validation_set : int
            The size of validation data set.
            Default 256.
        patience : int
            Stopping criterion for the training.
            Default 100.
        alpha : float
            The regularization parameter.
            Default "auto".

        Return
        ------
        self : object returns itself for convenience
        """
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        self.prepare_whitened_forward(noise_cov)
        n_channels, n_dipoles = self.leadfield.shape

        # Store simulation config
        self.simulation_config = simulation_config

        # Store Parameters
        # Architecture
        self.n_lstm_units = n_lstm_units
        self.n_dense_units = n_dense_units
        self.activation_function = activation_function
        # Training
        self.batch_size = self.simulation_config.batch_size
        self.epochs = epochs
        self.learning_rate = learning_rate
        self.loss = loss
        self.size_validation_set = size_validation_set
        self.patience = patience
        # Training Data (from simulation_config)
        self.n_timepoints = self.simulation_config.n_timepoints
        self.n_sources = self.simulation_config.n_sources
        self.n_orders = self.simulation_config.n_orders
        self.batch_repetitions = self.simulation_config.batch_repetitions
        self.snr_range = self.simulation_config.snr_range
        self.add_forward_error = self.simulation_config.add_forward_error
        self.forward_error = self.simulation_config.forward_error
        self.inter_source_correlation = self.simulation_config.inter_source_correlation
        self.correlation_mode = self.simulation_config.correlation_mode
        self.noise_color_coeff = self.simulation_config.noise_color_coeff

        # Inference
        logger.info("Create Generator:..")
        self.create_generator()
        logger.info("Build Model:..")
        self.build_model()
        logger.info("Train Model:..")
        self.train_model()

        self.inverse_operators = []
        return self

    def apply_inverse_operator(self, mne_obj) -> mne.SourceEstimate:
        """Apply the inverse operator.

        Parameters
        ----------
        mne_obj : [mne.Evoked, mne.Epochs, mne.io.Raw]
            The MNE data object.

        Return
        ------
        stc : mne.SourceEstimate
            The mne Source Estimate object.
        """
        data = self.unpack_data_obj(mne_obj)
        self.validate_operator_data_compatibility(data)
        data = self._sensor_transform @ data

        source_mat = self.apply_model(data)
        stc = self.source_to_object(source_mat)

        return stc

    def apply_model(self, data) -> np.ndarray:
        """Compute the inverse solution of the M/EEG data.

        Parameters
        ----------
        data : numpy.ndarray
            The M/EEG data matrix.

        Return
        ------
        x_hat : numpy.ndarray
            The source esimate.

        """
        y = deepcopy(data)
        y -= y.mean(axis=0)
        y /= np.linalg.norm(y, axis=0)
        y /= abs(y).max()

        n_channels, n_times = y.shape

        # Add empty batch and (color-) channel dimension
        y = y.T[np.newaxis]
        # Predict source(s)
        assert self.model is not None
        self.model.eval()
        device = self.device or get_torch_device()
        with torch.no_grad():
            source_pred = (
                self.model(torch.as_tensor(y, dtype=torch.float32, device=device))
                .detach()
                .cpu()
                .numpy()
            )
        source_pred = np.swapaxes(source_pred, 1, 2)  # (batch, dipoles, time)

        # Rescale sources
        y_original = deepcopy(data)
        y_original = y_original[np.newaxis]
        source_pred_scaled = rescale_sources(self.leadfield, source_pred[0], y_original)

        return source_pred_scaled

    def train_model(
        self,
    ):
        """Train the neural network model."""
        if self.model is None:
            raise RuntimeError("Model not initialized. Call build_model() first.")
        if self.optimizer is None:
            raise RuntimeError("Optimizer not initialized. Call build_model() first.")

        loss_fn = loss_from_name(self.loss)
        device = self.device or get_torch_device()

        # Get Validation data from generator
        x_val, y_val = next(self.generator)
        x_val = x_val[: self.size_validation_set]
        y_val = y_val[: self.size_validation_set]
        x_val_t = torch.as_tensor(x_val, dtype=torch.float32, device=device)
        y_val_t = torch.as_tensor(y_val, dtype=torch.float32, device=device)

        history: dict[str, list[float]] = {"loss": [], "val_loss": []}
        best_val = float("inf")
        best_state = None
        patience_left = int(self.patience)

        for _epoch in range(int(self.epochs)):
            self.model.train()
            running = 0.0
            for _step in range(int(self.batch_repetitions)):
                x_batch, y_batch = next(self.generator)
                x_t = torch.as_tensor(x_batch, dtype=torch.float32, device=device)
                y_t = torch.as_tensor(y_batch, dtype=torch.float32, device=device)

                self.optimizer.zero_grad(set_to_none=True)
                y_pred = self.model(x_t)
                loss = loss_fn(y_pred, y_t)
                loss.backward()
                self.optimizer.step()
                running += float(loss.detach().cpu().item())

            train_loss = running / float(self.batch_repetitions)

            self.model.eval()
            with torch.no_grad():
                val_loss = float(loss_fn(self.model(x_val_t), y_val_t).cpu().item())

            history["loss"].append(train_loss)
            history["val_loss"].append(val_loss)

            if val_loss < best_val:
                best_val = val_loss
                best_state = deepcopy(self.model.state_dict())
                patience_left = int(self.patience)
            else:
                patience_left -= 1
                if patience_left <= 0:
                    break

        if best_state is not None:
            self.model.load_state_dict(best_state)
        self.model.eval()
        self.history = history

    def build_model(
        self,
    ):
        """Build the neural network model."""
        n_channels, n_dipoles = self.leadfield.shape
        self.device = get_torch_device()
        self.model = _LSTMNet(
            n_channels,
            n_dipoles,
            n_dense_units=int(self.n_dense_units),
            n_lstm_units=int(self.n_lstm_units),
            activation_function=str(self.activation_function),
        ).to(self.device)
        self.optimizer = torch.optim.Adam(
            self.model.parameters(), lr=float(self.learning_rate)
        )
        logger.info(
            "Total number of trainable parameters: %d",
            count_trainable_parameters(self.model),
        )

    def create_generator(
        self,
    ):
        """Create the data generator used for the simulations."""
        # Create SimulationGenerator using the config
        sim_gen = SimulationGenerator(self.forward, config=self.simulation_config)

        # Wrap the generator to transpose and return (x, y) in the right format for LSTM
        def wrapped_generator():
            for x, y, _info in sim_gen.generate():
                # Transpose from (batch, channels, timepoints) to (batch, timepoints, channels)
                x = np.swapaxes(x, 1, 2)
                # Transpose from (batch, dipoles, timepoints) to (batch, timepoints, dipoles)
                y = np.swapaxes(y, 1, 2)

                yield x, y

        self.generator = wrapped_generator()



"""Shared Sparse Bayesian Learning (SBL) iteration engine.

This module extracts the common EM iteration loop from the Champagne-family
solvers (Champagne, FlexChampagne, OmniChampagne) into a reusable function.

Champagne (standard rules), FlexChampagne, and OmniChampagne use this engine
directly.  Complex Champagne variants (AR-EM, TEM, noise learning) retain
their own loops.

Typical usage::

    from invert.solvers.bayesian.sbl_core import sbl_iterate, GammaUpdateRule

    result = sbl_iterate(
        L=leadfield,
        Y=data,
        noise_cov=noise_cov,
        update_rule="mackay",
    )
    active_set = result.active_set
    gammas = result.gammas
"""

from __future__ import annotations

import dataclasses
from collections.abc import Callable

import numpy as np
from scipy.linalg import cho_factor, cho_solve


def _cholesky_inv(A):
    """Fast inverse for symmetric positive (semi-)definite matrices.

    Falls back to np.linalg.inv if Cholesky fails (e.g. near-singular or
    not quite PSD due to numerical noise).
    """
    n = A.shape[0]
    try:
        L, low = cho_factor(A)
        return cho_solve((L, low), np.eye(n))
    except np.linalg.LinAlgError:
        # Try with small diagonal loading proportional to trace
        eps = max(1e-10 * abs(np.trace(A)) / n, 1e-20)
        try:
            L, low = cho_factor(A + eps * np.eye(n))
            return cho_solve((L, low), np.eye(n))
        except np.linalg.LinAlgError:
            return np.linalg.inv(A + eps * np.eye(n))


@dataclasses.dataclass
class SBLResult:
    """Result of an SBL iteration run."""

    active_set: np.ndarray
    """Indices of surviving (non-pruned) atoms."""
    gammas: np.ndarray
    """Gamma values for surviving atoms."""
    mu_x: np.ndarray
    """Posterior mean for surviving atoms, shape (n_active, n_times)."""
    Sigma_y_inv: np.ndarray
    """Inverse model covariance at convergence."""
    loss: float
    """Final negative log-likelihood."""
    n_iters: int
    """Number of iterations executed."""


# ---------------------------------------------------------------------------
# Built-in gamma update rules
# ---------------------------------------------------------------------------


def _gamma_mackay(
    mu_x: np.ndarray,
    gammas: np.ndarray,
    z_diag: np.ndarray,
    n_times: int,
) -> np.ndarray:
    upper = np.mean(mu_x**2, axis=1)
    return upper / (gammas * z_diag + 1e-20)


def _gamma_convexity(
    mu_x: np.ndarray,
    gammas: np.ndarray,
    z_diag: np.ndarray,
    n_times: int,
) -> np.ndarray:
    upper = np.mean(mu_x**2, axis=1)
    return np.sqrt(upper / (z_diag + 1e-20))


def _gamma_em(
    mu_x: np.ndarray,
    gammas: np.ndarray,
    z_diag: np.ndarray,
    n_times: int,
) -> np.ndarray:
    diag_sigma_x = gammas - gammas**2 * z_diag
    return diag_sigma_x + np.mean(mu_x**2, axis=1)


GammaUpdateRule = Callable[[np.ndarray, np.ndarray, np.ndarray, int], np.ndarray]
"""Signature: (mu_x, gammas, z_diag, n_times) -> gammas_new."""

_BUILTIN_RULES: dict[str, GammaUpdateRule] = {
    "mackay": _gamma_mackay,
    "convexity": _gamma_convexity,
    "mm": _gamma_convexity,
    "em": _gamma_em,
}


def _resolve_update_rule(rule: str | GammaUpdateRule) -> GammaUpdateRule:
    if callable(rule):
        return rule
    key = str(rule).lower()
    if key not in _BUILTIN_RULES:
        raise ValueError(
            f"Unknown update rule {rule!r}. "
            f"Supported: {sorted(_BUILTIN_RULES.keys())} or a callable."
        )
    return _BUILTIN_RULES[key]


# ---------------------------------------------------------------------------
# Main iteration engine
# ---------------------------------------------------------------------------


def sbl_iterate(
    L: np.ndarray,
    Y: np.ndarray,
    noise_cov: np.ndarray,
    update_rule: str | GammaUpdateRule = "mackay",
    *,
    max_iter: int = 2000,
    prune: bool = True,
    pruning_thresh: float = 1e-3,
    conv_crit: float = 1e-8,
    inverse_fn: Callable[[np.ndarray], np.ndarray] = _cholesky_inv,
) -> SBLResult:
    """Run a generic SBL iteration loop.

    Parameters
    ----------
    L : ndarray (n_chans, n_atoms)
        Leadfield (or extended dictionary).
    Y : ndarray (n_chans, n_times)
        Data matrix.
    noise_cov : ndarray (n_chans, n_chans)
        Noise covariance.
    update_rule : str or callable
        Gamma update rule.  Built-in: ``"mackay"``, ``"convexity"``/``"mm"``,
        ``"em"``.  Or a callable with signature
        ``(mu_x, gammas, z_diag, n_times) -> gammas_new``.
    max_iter : int
        Maximum EM iterations.
    prune : bool
        Whether to prune inactive atoms each iteration.
    pruning_thresh : float
        Relative threshold: atoms with gamma < pruning_thresh * max(gamma)
        are pruned.
    conv_crit : float
        Relative change in loss below which iteration stops.
    inverse_fn : callable
        Function to invert a matrix, e.g. ``np.linalg.inv`` or a
        regularized variant.

    Returns
    -------
    SBLResult
        Dataclass with ``active_set``, ``gammas``, ``mu_x``,
        ``Sigma_y_inv``, ``loss``, ``n_iters``.
    """
    rule_fn = _resolve_update_rule(update_rule)
    n_chans, n_atoms = L.shape
    n_times = Y.shape[1]

    gammas = np.ones(n_atoms)
    active_set = np.arange(n_atoms)
    L_act = L.copy()

    # Initial posterior
    Sigma_y = noise_cov + (L_act * gammas) @ L_act.T
    Sigma_y = 0.5 * (Sigma_y + Sigma_y.T)
    Sigma_y_inv = inverse_fn(Sigma_y)
    SiL = Sigma_y_inv @ L_act  # cached for z_diag and mu_x
    mu_x = (SiL.T @ Y) * gammas[:, None]

    # Compute initial loss so we have a valid value even if the loop breaks
    # on iteration 0 (e.g. all gammas go to zero immediately).
    _data_fit_init = float(np.trace(Sigma_y_inv @ Y @ Y.T) / n_times)
    with np.errstate(divide="ignore", over="ignore", invalid="ignore"):
        _sign_init, _log_det_init = np.linalg.slogdet(Sigma_y)
    if _sign_init <= 0 or not np.isfinite(_log_det_init):
        _log_det_init = np.finfo(float).max / 2
    final_loss = _data_fit_init + float(_log_det_init)

    loss_prev: float | None = None
    n_iters = 0

    for i_iter in range(max_iter):
        n_iters = i_iter + 1

        # Fisher information diagonal: diag(L.T @ Sigma_y_inv @ L)
        # SiL is cached from posterior recompute (or initial)
        z_diag = np.sum(L_act * SiL, axis=0)

        # Gamma update (rule-specific)
        gammas_new = rule_fn(mu_x, gammas, z_diag, n_times)

        # Cleanup
        gammas_new[~np.isfinite(gammas_new)] = 0.0
        gammas_new = np.maximum(gammas_new, 0.0)
        if float(np.linalg.norm(gammas_new)) == 0.0:
            break

        # Pruning
        if prune:
            thresh = pruning_thresh * float(gammas_new.max())
            keep = np.where(gammas_new > thresh)[0]
            if keep.size == 0:
                break
            active_set = active_set[keep]
            gammas = gammas_new[keep]
            L_act = L_act[:, keep]
        else:
            gammas = gammas_new

        # Recompute posterior after pruning
        Sigma_y = noise_cov + (L_act * gammas) @ L_act.T
        Sigma_y = 0.5 * (Sigma_y + Sigma_y.T)
        Sigma_y_inv = inverse_fn(Sigma_y)
        SiL = Sigma_y_inv @ L_act  # cached for z_diag in next iteration
        mu_x = (SiL.T @ Y) * gammas[:, None]

        # Loss: negative log-marginal-likelihood
        # slogdet is faster than eigvalsh for just the log-determinant
        data_fit = float(np.trace(Sigma_y_inv @ Y @ Y.T) / n_times)
        with np.errstate(divide="ignore", over="ignore", invalid="ignore"):
            sign, log_det = np.linalg.slogdet(Sigma_y)
        if sign <= 0 or not np.isfinite(log_det):
            log_det = np.finfo(float).max / 2
        loss = data_fit + float(log_det)
        final_loss = loss

        # Convergence check
        if loss_prev is not None:
            rel_change = (loss_prev - loss) / (abs(loss_prev) + 1e-20)
            if rel_change > 0 and rel_change < conv_crit:
                loss_prev = loss
                break
        loss_prev = loss

    return SBLResult(
        active_set=active_set,
        gammas=gammas,
        mu_x=mu_x,
        Sigma_y_inv=Sigma_y_inv,
        loss=final_loss,
        n_iters=n_iters,
    )



from dataclasses import dataclass
from typing import Literal, TypeAlias

import mne
import numpy as np
from scipy.linalg import eigh
from sklearn.covariance import OAS

ArrayLike = np.ndarray
MNEObj: TypeAlias = mne.io.BaseRaw | mne.Epochs | mne.Evoked | mne.EvokedArray


# ============================
# Utilities
# ============================


def _hann(n: int) -> np.ndarray:
    # symmetric Hann for perfect overlap-add with 50% step
    return 0.5 - 0.5 * np.cos(2 * np.pi * np.arange(n) / (n - 1))


def _overlap_add(n_times: int, win_len: int, step: int):
    starts = list(range(0, max(1, n_times - win_len + 1), step))
    if len(starts) == 0 or starts[-1] != n_times - win_len:
        starts.append(max(0, n_times - win_len))
    return starts


def _safe_shrinkage_cov(X: np.ndarray) -> np.ndarray:
    # X: (n_ch, n_t) demeaned
    if X.shape[1] < X.shape[0]:  # too few samples -> fallback to diagonal
        s = np.var(X, axis=1, ddof=1)
        return np.diag(np.maximum(s, 1e-12))
    oas = OAS(assume_centered=True).fit(X.T)
    return oas.covariance_


def _spd_from_eig(Q: np.ndarray, d: np.ndarray) -> np.ndarray:
    return Q @ np.diag(np.maximum(d, 1e-12)) @ Q.T


def _inv_sqrt_spd(Q: np.ndarray, d: np.ndarray) -> np.ndarray:
    return Q @ np.diag(1.0 / np.sqrt(np.maximum(d, 1e-12))) @ Q.T


# ============================
# Base class
# ============================


@dataclass
class DenoiserBase:
    sfreq: float | None = None
    window_size: float = 0.5  # seconds
    step_frac: float = 0.5  # overlap = 1 - step_frac
    use_windows: bool = True  # set False to process full signal

    def _ensure_sfreq(self, sfreq: float | None):
        if self.sfreq is None:
            if sfreq is None:
                raise ValueError(
                    "sfreq must be provided (or set on class) for NumPy input."
                )
            self.sfreq = sfreq

    # ---- public API on NumPy ----
    def run(self, X: ArrayLike, sfreq: float | None = None) -> ArrayLike:
        """
        X shape must be (n_channels, n_times).
        """
        self._ensure_sfreq(sfreq)
        X = np.asarray(X, dtype=float)
        assert X.ndim == 2, "X must be (n_channels, n_times)"
        n_ch, n_t = X.shape
        if not self.use_windows:
            return self._process_window(X)

        assert self.sfreq is not None
        win_len = max(8, int(round(self.window_size * self.sfreq)))
        step = max(1, int(round(win_len * self.step_frac)))
        if n_t < 2 * win_len:  # too short for sliding windows -> process once
            return self._process_window(X)

        w = _hann(win_len)
        starts = _overlap_add(n_t, win_len, step)
        Y = np.zeros_like(X)
        wsum = np.zeros(n_t)

        for s in starts:
            seg = X[:, s : s + win_len].copy()
            seg -= seg.mean(axis=1, keepdims=True)
            seg *= w  # apply window in time
            clean = self._process_window(seg)  # (n_ch, win_len)
            Y[:, s : s + win_len] += clean
            wsum[s : s + win_len] += w

        wsum = np.clip(wsum, 1e-12, None)
        Y /= wsum
        return Y

    # ---- public API for MNE ----
    def run_mne(self, obj: MNEObj):
        """
        Returns the same MNE type with data cleaned.
        """
        if isinstance(obj, mne.io.BaseRaw):
            self._ensure_sfreq(obj.info["sfreq"])
            data = obj.get_data()
            cleaned = self.run(data, sfreq=self.sfreq)
            return mne.io.RawArray(cleaned, obj.info)

        if isinstance(obj, mne.Epochs):
            self._ensure_sfreq(obj.info["sfreq"])
            X = obj.get_data()  # (n_ep, n_ch, n_t)
            cleaned = np.empty_like(X)
            for i in range(X.shape[0]):
                cleaned[i] = self.run(X[i], sfreq=self.sfreq)
            return mne.EpochsArray(cleaned, obj.info, events=obj.events, tmin=obj.tmin)

        if isinstance(obj, (mne.Evoked, mne.EvokedArray)):
            self._ensure_sfreq(obj.info["sfreq"])
            cleaned = self.run(obj.data, sfreq=self.sfreq)
            return mne.EvokedArray(cleaned, obj.info, tmin=obj.times[0])

        raise TypeError(f"Unsupported MNE object: {type(obj)}")

    # ---- to be implemented by subclasses ----
    def _process_window(self, Xw: ArrayLike) -> ArrayLike:
        raise NotImplementedError


# ============================
# Variant A: Leadfield-Guided Subspace Projection (LGSP)
# ============================


@dataclass
class LGSP(DenoiserBase):
    L: np.ndarray | None = None  # (m, n_sources)
    sigma_prior: Literal["identity"] = "identity"
    lambda_ref: float = 1e-3  # ridge on C_ref
    alpha_model: float = 0.2  # blend with scaled I: Cref'=(1-a)Cref+a*trace(Cref)/m*I
    rank: int | None = None  # keep r smallest generalized eigenvalues
    tau: float | None = None  # or threshold on generalized eigenvalues
    center: bool = True
    shrink_data: bool = True

    def _build_Cref(self, n_ch: int) -> np.ndarray:
        if self.L is None:
            # fall back: isotropic brain covariance
            Cref = np.eye(n_ch)
        else:
            L = self.L
            if self.sigma_prior == "identity":
                # C_ref = L L^T (scaled) + lambda I
                Cref = L @ L.T
        # scale & regularize
        tr = np.trace(Cref) / max(1, n_ch)
        Cref = (1 - self.alpha_model) * Cref + self.alpha_model * tr * np.eye(
            Cref.shape[0]
        )
        Cref = Cref + self.lambda_ref * np.eye(Cref.shape[0])
        return Cref

    def _gevd_brain_projector(
        self, Cx: np.ndarray, Cref: np.ndarray, r: int | None, tau: float | None
    ) -> np.ndarray:
        """
        Return the sensor-space projector P onto the 'brain-like' subspace
        spanned by generalized eigenvectors with the smallest eigenvalues.
        P is constructed as a B-orthogonal projector: P = V (V^T B V)^{-1} V^T B
        where columns of V are the chosen generalized eigenvectors and B=Cref.
        """
        # Solve symmetric definite generalized eigenproblem: Cx v = eta Cref v
        # eigh returns eigenvalues in ascending order for symmetric problems.
        eta, V = eigh(Cx, Cref, turbo=True)
        # choose modes (smallest eta are more brain-like)
        if r is None and tau is None:
            r = max(1, min(Cx.shape[0] - 1, int(round(0.7 * Cx.shape[0]))))
        if tau is not None:
            keep = np.where(eta <= tau)[0]
            if keep.size == 0:
                keep = np.arange(min(1, len(eta)))  # type: ignore[arg-type, type-var]
        else:
            keep = np.arange(min(r, len(eta)))  # type: ignore[arg-type, type-var]
        Vr = V[:, keep]  # (m, k)
        assert Cref is not None
        B = Cref
        G = Vr.T @ B @ Vr  # (k, k)
        Ginv = np.linalg.pinv(G)
        P = Vr @ Ginv @ Vr.T @ B  # (m, m) projector in B-metric
        # P is idempotent in B-metric: P^2 ≈ P (numerically)
        return P

    def _process_window(self, Xw: ArrayLike) -> ArrayLike:
        # Xw: (n_ch, n_t), mean already subtracted & windowed
        n_ch, _ = Xw.shape
        if self.center:
            Xw = Xw - Xw.mean(axis=1, keepdims=True)

        Cref = self._build_Cref(n_ch)
        # Shrinkage covariance of data (robust in low-sample windows)
        Cx = _safe_shrinkage_cov(Xw) if self.shrink_data else np.cov(Xw)

        P = self._gevd_brain_projector(Cx, Cref, self.rank, self.tau)
        return P @ Xw


# ============================
# Variant B: Source-space Reprojection & Blend (SRB)
# ============================


@dataclass
class SRB(DenoiserBase):
    L: np.ndarray | None = None  # (m, n_sources) required for meaningful use
    mu: float = 1e-1  # Tikhonov on sources (||S||^2)
    beta: float = 0.5  # blend factor: X_clean = beta*L*W*X + (1-beta)*X
    center: bool = True
    cache_inverse: bool = True

    def __post_init__(self):
        self._W = None  # cached inverse

    def _compute_inverse(self, L: np.ndarray) -> np.ndarray:
        # W = (L^T L + mu I)^{-1} L^T
        LtL = L.T @ L
        n_src = LtL.shape[0]
        A = LtL + self.mu * np.eye(n_src)
        W = np.linalg.solve(A, L.T)
        return W

    def _ensure_inverse(self, n_ch: int):
        if self.L is None:
            # fall back: identity projection (no-op)
            self._W = None
            return
        if (self._W is None) or (self._W.shape[1] != self.L.shape[0]):
            self._W = self._compute_inverse(self.L)

    def _process_window(self, Xw: ArrayLike) -> ArrayLike:
        if self.center:
            Xw = Xw - Xw.mean(axis=1, keepdims=True)
        n_ch, _ = Xw.shape
        self._ensure_inverse(n_ch)
        if self._W is None:  # no leadfield provided
            return Xw
        S = self._W @ Xw  # (n_src, n_t)
        Xb = self.L @ S  # (n_ch, n_t) projection to brain-span
        return self.beta * Xb + (1.0 - self.beta) * Xw


# ============================
# Usage Examples
# ============================

# -- NumPy data (n_ch, n_times) --
# sfreq must be provided if not set on class
# L: (n_ch, n_src) leadfield

# LGSP example:
# lgsp = LGSP(L=L, window_size=0.5, step_frac=0.5, rank=3, lambda_ref=1e-3, alpha_model=0.3)
# X_clean = lgsp.run(X, sfreq=256)

# SRB example:
# srb = SRB(L=L, mu=1e-1, beta=0.6, use_windows=False)  # SRB often fine without windowing
# X_clean = srb.run(X, sfreq=256)

# -- MNE objects --
# lgsp_mne = lgsp.run_mne(raw)         # returns Raw
# srb_mne  = srb.run_mne(epochs)       # returns Epochs



import numpy as np


def _weighted_mne_expected(leadfield, y, weights, alpha):
    """Reference implementation used for adapter tests."""
    weights = np.maximum(np.asarray(weights, dtype=float), 1e-12)
    LW = leadfield * weights[np.newaxis, :]
    A = LW @ leadfield.T + alpha * np.eye(leadfield.shape[0])
    z = np.linalg.solve(A, y)
    return weights * (leadfield.T @ z)


class TestFocussAdapter:
    def test_single_iter_matches_weighted_mne(
        self, forward_model, simulated_evoked, simulated_stc
    ):
        """With max_iter=1, FOCUSS should equal a single weighted MNE update."""
        from invert.adapters.focuss import focuss

        alpha = 0.05
        stc_out = focuss(
            simulated_stc, simulated_evoked, forward_model, alpha=alpha, max_iter=1
        )

        L = forward_model["sol"]["data"]
        y = simulated_evoked.data[:, 0]
        x0 = simulated_stc.data[:, 0]
        expected = _weighted_mne_expected(L, y, np.abs(x0), alpha)

        np.testing.assert_allclose(stc_out.data[:, 0], expected, atol=1e-10, rtol=1e-10)


class TestSmooth:
    def test_keeps_prominent_without_self_edges(self):
        from invert.adapters.focuss import smooth

        adjacency = np.array(
            [
                [0, 1, 0],
                [1, 0, 1],
                [0, 1, 0],
            ],
            dtype=int,
        )
        x = np.array([10.0, 0.0, 0.0])
        x_smooth, sparsity = smooth(x, adjacency, percentile=0.5)

        assert sparsity == 1
        assert x_smooth[0] != 0
        assert x_smooth[2] == 0

    def test_zero_input_is_stable(self):
        from invert.adapters.focuss import smooth

        adjacency = np.zeros((3, 3), dtype=int)
        x = np.zeros(3)
        x_smooth, sparsity = smooth(x, adjacency, percentile=0.5)
        np.testing.assert_allclose(x_smooth, 0.0, atol=0.0)
        assert sparsity == 0


class TestOtherAdapters:
    def test_s_focuss_runs_and_is_finite(
        self, forward_model, simulated_evoked, simulated_stc
    ):
        from invert.adapters.focuss import s_focuss

        stc_out = s_focuss(
            simulated_stc,
            simulated_evoked,
            forward_model,
            alpha=0.05,
            percentile=0.5,
            max_iter=2,
        )
        assert stc_out.data.shape == simulated_stc.data.shape
        assert np.all(np.isfinite(stc_out.data))

    def test_stampc_does_not_mutate_forward(
        self, forward_model, simulated_evoked, simulated_stc
    ):
        from invert.adapters.stamp import stampc

        leadfield_before = forward_model["sol"]["data"].copy()
        _ = stampc(
            simulated_stc,
            simulated_evoked,
            forward_model,
            max_iter=2,
            K=1,
            n_orders=0,
        )
        np.testing.assert_allclose(
            forward_model["sol"]["data"], leadfield_before, atol=0.0
        )

    def test_stampc_runs_with_orders_and_zero_seed(
        self, forward_model, simulated_evoked, simulated_stc
    ):
        from invert.adapters.stamp import stampc

        stc_zero = simulated_stc.copy()
        stc_zero.data[:] = 0.0

        stc_out = stampc(
            stc_zero,
            simulated_evoked,
            forward_model,
            max_iter=3,
            K=2,
            rv_thresh=0.1,
            n_orders=1,
        )

        assert stc_out.data.shape == simulated_stc.data.shape
        assert np.all(np.isfinite(stc_out.data))



from __future__ import annotations

import inspect
from collections.abc import Mapping, Sequence
from functools import cache
from typing import Any

from invert.solver_registry import (
    get_solver_class,
    get_solver_spec,
    iter_solver_specs,
)

RECOMMENDED_DEPTH_BY_CATEGORY: dict[str, float] = {
    "minimum_norm": 0.5,
    "loreta": 0.0,
    "bayesian": 0.0,
    "music": 0.5,
}

RECOMMENDED_DEPTH_BY_SOLVER: dict[str, float] = {
    "backus-gilbert": 0.0,
    "gbf": 1.0,
    "gft-l1": 0.0,
    "mne": 1.0,
    "s-map": 0.0,
    "tv": 0.0,
}


def _infer_category(module_path: str, solver_id: str) -> str:
    parts = module_path.split(".")
    if len(parts) < 3 or parts[0] != "invert" or parts[1] != "solvers":
        return "other"
    directory = parts[2]
    if directory == "beamformers":
        return "beamformer"
    if directory == "bayesian":
        return "bayesian"
    if directory == "minimum_norm":
        if "loreta" in solver_id:
            return "loreta"
        return "minimum_norm"
    if directory == "music":
        return "music"
    if directory == "matching_pursuit":
        return "matching_pursuit"
    if directory == "neural_networks":
        return "neural_networks"
    return "other"


@cache
def infer_depth_strategy(solver_name: str) -> str | None:
    """Infer how depth should be set for a solver.

    Returns one of:
    - ``"base"``: BaseSolver depth weighting (``use_depth_weighting`` + ``depth_weighting``)
    - ``"dspm-depth"``: dSPM-style depth prior via ``depth`` with ``use_depth_weighting=True``
    - ``"laura-depth"``: LAURA-specific ``depth_weight``
    - ``"make-depth-weighting"``: make-time ``depth_weighting`` argument
    - ``None``: no detectable depth control
    """
    spec = get_solver_spec(solver_name)
    solver_cls = get_solver_class(spec.solver_id)

    init_sig = inspect.signature(solver_cls.__init__)
    make_sig = inspect.signature(solver_cls.make_inverse_operator)
    init_params = init_sig.parameters
    make_params = make_sig.parameters

    if "depth_weight" in init_params:
        return "laura-depth"
    if "depth" in init_params:
        return "dspm-depth"
    if "use_depth_weighting" in init_params or "depth_weighting" in init_params:
        return "base"

    accepts_init_kwargs = any(
        p.kind == inspect.Parameter.VAR_KEYWORD for p in init_params.values()
    )
    if accepts_init_kwargs:
        try:
            solver = solver_cls(**dict(spec.default_kwargs))
        except Exception:
            solver = None

        if solver is not None and bool(getattr(solver, "prep_leadfield", True)):
            return "base"

    if "depth_weighting" in make_params:
        return "make-depth-weighting"
    return None


def build_depth_params_for_solver(solver_name: str, depth: float) -> dict[str, Any]:
    """Build benchmark ``solver_params`` entries for a solver at one depth."""
    strategy = infer_depth_strategy(solver_name)
    if strategy is None:
        return {}
    if strategy == "laura-depth":
        return {"init__depth_weight": float(depth)}
    if strategy == "dspm-depth":
        return {"init__use_depth_weighting": True, "init__depth": float(depth)}
    if strategy == "make-depth-weighting":
        return {"make__depth_weighting": float(depth)}
    return {"init__use_depth_weighting": True, "init__depth_weighting": float(depth)}


def scan_depth_effective_solvers(
    *,
    solver_names: Sequence[str] | None = None,
    include_unavailable: bool = False,
) -> dict[str, dict[str, str]]:
    """Return solver -> metadata for solvers where depth has a detectable effect."""
    if solver_names is None:
        names = [
            spec.solver_id
            for spec in iter_solver_specs(
                include_internal=False,
                include_unavailable=include_unavailable,
            )
        ]
    else:
        names = [get_solver_spec(name).solver_id for name in solver_names]

    out: dict[str, dict[str, str]] = {}
    for solver_name in names:
        strategy = infer_depth_strategy(solver_name)
        if strategy is None:
            continue
        spec = get_solver_spec(solver_name)
        out[solver_name] = {
            "category": _infer_category(spec.module_path, spec.solver_id),
            "strategy": strategy,
            "module_path": spec.module_path,
        }
    return out


def build_depth_solver_params(
    *,
    solvers: Sequence[str],
    depth_by_category: Mapping[str, float] | None,
    depth_by_solver: Mapping[str, float] | None = None,
    base_solver_params: Mapping[str, Mapping[str, Any]] | None = None,
) -> dict[str, dict[str, Any]]:
    """Compose per-solver depth params from category+solver defaults + overrides.

    Precedence:
    1) category depth defaults
    2) per-solver depth overrides
    3) explicit keys in ``base_solver_params``
    """
    merged: dict[str, dict[str, Any]] = {}
    selected_solver_ids = {get_solver_spec(name).solver_id for name in solvers}

    if depth_by_category:
        for solver_name in solvers:
            sid = get_solver_spec(solver_name).solver_id
            spec = get_solver_spec(sid)
            category = _infer_category(spec.module_path, sid)
            if category not in depth_by_category:
                continue
            depth = float(depth_by_category[category])
            depth_params = build_depth_params_for_solver(sid, depth)
            if depth_params:
                merged.setdefault(sid, {}).update(depth_params)

    if depth_by_solver:
        for solver_name, depth in depth_by_solver.items():
            sid = get_solver_spec(solver_name).solver_id
            if sid not in selected_solver_ids:
                continue
            depth_params = build_depth_params_for_solver(sid, float(depth))
            if depth_params:
                merged.setdefault(sid, {}).update(depth_params)

    if base_solver_params:
        for solver_name, params in base_solver_params.items():
            sid = get_solver_spec(solver_name).solver_id
            merged.setdefault(sid, {}).update(dict(params))

    return merged
